import enum

#https=//docs.microsoft.com/en-us/openspecs/windows_protocols/ms-cifs/32b5d4b7-d90b-483f-ad6a-003fd110f0ec
class SMBCommand(enum.Enum):
	SMB_COM_CREATE_DIRECTORY = 0x00
	SMB_COM_DELETE_DIRECTORY = 0x01
	SMB_COM_OPEN = 0x02
	SMB_COM_CREATE = 0x03
	SMB_COM_CLOSE = 0x04
	SMB_COM_FLUSH = 0x05
	SMB_COM_DELETE = 0x06
	SMB_COM_RENAME = 0x07
	SMB_COM_QUERY_INFORMATION = 0x08
	SMB_COM_SET_INFORMATION = 0x09
	SMB_COM_READ = 0x0A
	SMB_COM_WRITE = 0x0B
	SMB_COM_LOCK_BYTE_RANGE = 0x0C
	SMB_COM_UNLOCK_BYTE_RANGE = 0x0D
	SMB_COM_CREATE_TEMPORARY = 0x0E
	SMB_COM_CREATE_NEW = 0x0F
	SMB_COM_CHECK_DIRECTORY = 0x10
	SMB_COM_PROCESS_EXIT = 0x11
	SMB_COM_SEEK = 0x12
	SMB_COM_LOCK_AND_READ = 0x13
	SMB_COM_WRITE_AND_UNLOCK = 0x14
	SMB_COM_READ_RAW = 0x1A
	SMB_COM_READ_MPX = 0x1B
	SMB_COM_READ_MPX_SECONDARY = 0x1C
	SMB_COM_WRITE_RAW = 0x1D
	SMB_COM_WRITE_MPX = 0x1E
	SMB_COM_WRITE_MPX_SECONDARY = 0x1F
	SMB_COM_WRITE_COMPLETE = 0x20
	SMB_COM_QUERY_SERVER = 0x21
	SMB_COM_SET_INFORMATION2 = 0x22
	SMB_COM_QUERY_INFORMATION2 = 0x23
	SMB_COM_LOCKING_ANDX = 0x24
	SMB_COM_TRANSACTION = 0x25
	SMB_COM_TRANSACTION_SECONDARY = 0x26
	SMB_COM_IOCTL = 0x27
	SMB_COM_IOCTL_SECONDARY = 0x28
	SMB_COM_COPY = 0x29
	SMB_COM_MOVE = 0x2A
	SMB_COM_ECHO = 0x2B
	SMB_COM_WRITE_AND_CLOSE = 0x2C
	SMB_COM_OPEN_ANDX = 0x2D
	SMB_COM_READ_ANDX = 0x2E
	SMB_COM_WRITE_ANDX = 0x2F
	SMB_COM_NEW_FILE_SIZE = 0x30
	SMB_COM_CLOSE_AND_TREE_DISC = 0x31
	SMB_COM_TRANSACTION2 = 0x32
	SMB_COM_TRANSACTION2_SECONDARY = 0x33
	SMB_COM_FIND_CLOSE2 = 0x34
	SMB_COM_FIND_NOTIFY_CLOSE = 0x35
	SMB_COM_TREE_CONNECT = 0x70
	SMB_COM_TREE_DISCONNECT = 0x71
	SMB_COM_NEGOTIATE = 0x72
	SMB_COM_SESSION_SETUP_ANDX = 0x73
	SMB_COM_LOGOFF_ANDX = 0x74
	SMB_COM_TREE_CONNECT_ANDX = 0x75
	SMB_COM_SECURITY_PACKAGE_ANDX = 0x7E
	SMB_COM_QUERY_INFORMATION_DISK = 0x80
	SMB_COM_SEARCH = 0x81
	SMB_COM_FIND = 0x82
	SMB_COM_FIND_UNIQUE = 0x83
	SMB_COM_FIND_CLOSE = 0x84
	SMB_COM_NT_TRANSACT = 0xA0
	SMB_COM_NT_TRANSACT_SECONDARY = 0xA1
	SMB_COM_NT_CREATE_ANDX = 0xA2
	SMB_COM_NT_CANCEL = 0xA4
	SMB_COM_NT_RENAME = 0xA5
	SMB_COM_OPEN_PRINT_FILE = 0xC0
	SMB_COM_WRITE_PRINT_FILE = 0xC1
	SMB_COM_CLOSE_PRINT_FILE = 0xC2
	SMB_COM_GET_PRINT_QUEUE  = 0xC3
	SMB_COM_READ_BULK  = 0xD8
	SMB_COM_WRITE_BULK = 0xD9
	SMB_COM_WRITE_BULK_DATA  = 0xDA
	SMB_COM_INVALID = 0xFE
	SMB_COM_NO_ANDX_COMMAND = 0xFF


import io
import enum

from aiosmb.wintypes.fscc.structures.fileinfoclass import *

class QueryDirectoryFlag(enum.IntFlag):
	SMB2_RESTART_SCANS = 0x01 #The server MUST restart the enumeration from the beginning as specified in section 3.3.5.18.
	SMB2_RETURN_SINGLE_ENTRY = 0x02 #The server MUST only return the first entry of the search results.
	SMB2_INDEX_SPECIFIED = 0x04 #The server SHOULD<64> return entries beginning at the byte number specified by FileIndex.
	SMB2_REOPEN = 0x10 #The server MUST restart the enumeration from the beginning, and the search pattern MUST be changed to the provided value.

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/10906442-294c-46d3-8515-c277efe1f752
class QUERY_DIRECTORY_REQ:
	def __init__(self):
		self.StructureSize = 33
		self.FileInformationClass  = None
		self.Flags  = None
		self.FileIndex  = None
		self.FileId  = None
		self.FileNameOffset  = 0
		self.FileNameLength  = 0
		self.OutputBufferLength  = 65535
		
		self.Buffer = None
		
		#high-level param
		self.FileName = None #this is actually a search pattern, but referred to as filename
		
	def to_bytes(self):
		if self.FileName:
			self.Buffer = self.FileName.encode('utf-16-le')
			self.FileNameOffset = 64 + 2+1+1+4+16+2+2+4
			self.FileNameLength = len(self.Buffer)
			
		t  = self.StructureSize.to_bytes(2, byteorder='little', signed = False)
		t += self.FileInformationClass.value.to_bytes(1, byteorder='little', signed = False)
		t += self.Flags.to_bytes(1, byteorder='little', signed = False)
		t += self.FileIndex.to_bytes(4, byteorder='little', signed = False)
		t += self.FileId.to_bytes(16, byteorder='little', signed = False)
		t += self.FileNameOffset.to_bytes(2, byteorder='little', signed = False)
		t += self.FileNameLength.to_bytes(2, byteorder='little', signed = False)
		t += self.OutputBufferLength.to_bytes(4, byteorder='little', signed = False)
		
		if self.Buffer:
			t += self.Buffer
		return t

	@staticmethod
	def from_bytes(bbuff):
		return QUERY_DIRECTORY_REQ.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = QUERY_DIRECTORY_REQ()
		msg.StructureSize   = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 33
		msg.FileInformationClass  = FileInformationClass(int.from_bytes(buff.read(1), byteorder='little'))
		msg.Flags   = QueryDirectoryFlag(int.from_bytes(buff.read(1), byteorder='little'))
		msg.FileIndex    = int.from_bytes(buff.read(4), byteorder = 'little')
		msg.FileId    = int.from_bytes(buff.read(16), byteorder = 'little')
		msg.FileNameOffset = int.from_bytes(buff.read(2), byteorder = 'little')
		msg.FileNameLength = int.from_bytes(buff.read(2), byteorder = 'little')
		msg.OutputBufferLength = int.from_bytes(buff.read(4), byteorder = 'little')

		if msg.FileNameLength > 0:
			buff.seek(msg.FileNameOffset, io.SEEK_SET)
			msg.FileName = buff.read(msg.FileNameLength).deocde('utf-16-le')

		return msg

	def __repr__(self):
		t = '==== SMB2 QUERY DIRECTORY REQ ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'FileInformationClass: %s\r\n' % self.FileInformationClass
		t += 'Flags: %s\r\n' % self.Flags
		t += 'FileIndex: %s\r\n' % self.FileIndex
		t += 'FileId: %s\r\n' % self.FileId
		t += 'FileNameOffset: %s\r\n' % self.FileNameOffset
		t += 'FileNameLength: %s\r\n' % self.FileNameLength
		t += 'OutputBufferLength: %s\r\n' % self.OutputBufferLength
		t += 'FileName: %s\r\n' % self.FileName
		return t
		
		
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/4f75351b-048c-4a0c-9ea3-addd55a71956
class QUERY_DIRECTORY_REPLY:
	def __init__(self):
		self.StructureSize = 9
		self.OutputBufferOffset  = None
		self.OutputBufferLength  = None
		self.Buffer = None
		
		#high-level param
		self.Data = None
		
	def to_bytes(self):
		pass

	@staticmethod
	def from_bytes(bbuff):
		return QUERY_DIRECTORY_REPLY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = QUERY_DIRECTORY_REPLY()
		msg.StructureSize   = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 9
		msg.OutputBufferOffset  = int.from_bytes(buff.read(2), byteorder='little')
		msg.OutputBufferLength  = int.from_bytes(buff.read(4), byteorder='little')
		
		if msg.OutputBufferLength > 0:
			buff.seek(msg.OutputBufferOffset, io.SEEK_SET)
			msg.Data= buff.read(msg.OutputBufferLength)

		return msg

	def __repr__(self):
		t = '==== SMB2 QUERY INFO REQ ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'OutputBufferOffset: %s\r\n' % self.OutputBufferOffset
		t += 'OutputBufferLength: %s\r\n' % self.OutputBufferLength
		t += 'Data: %s\r\n' % self.Data
		return t


#https://www.ietf.org/rfc/rfc1928.txt
#https://tools.ietf.org/html/rfc1929
#https://tools.ietf.org/html/rfc1961

import io
import enum
import ipaddress
import socket
import asyncio

async def readexactly_or_exc(reader, n, timeout = None):
	"""
	Helper function to read exactly N amount of data from the wire.
	:param reader: The reader object
	:type reader: asyncio.StreamReader
	:param n: The maximum amount of bytes to read.
	:type n: int
	:param timeout: Time in seconds to wait for the reader to return data
	:type timeout: int
	:return: bytearray
	"""
	temp = await asyncio.gather(*[asyncio.wait_for(reader.readexactly(n), timeout = timeout)], return_exceptions=True)
	if isinstance(temp[0], bytes):
		return temp[0]
	else:
		raise temp[0]


async def read_or_exc(reader, n, timeout = None):
	"""
	Helper function to read N amount of data from the wire.
	:param reader: The reader object
	:type reader: asyncio.StreamReader
	:param n: The maximum amount of bytes to read. BEWARE: this only sets an upper limit of the data to be read
	:type n: int
	:param timeout: Time in seconds to wait for the reader to return data
	:type timeout: int
	:return: bytearray
	"""

	temp = await asyncio.gather(*[asyncio.wait_for(reader.read(n), timeout = timeout)], return_exceptions=True)
	if isinstance(temp[0], bytes):
		return temp[0]
	else:
		raise temp[0]


class SOCKS5ServerState(enum.Enum):
	NEGOTIATION = 0
	NOT_AUTHENTICATED = 1
	REQUEST = 3 
	RELAYING = 4


class SOCKS5Method(enum.Enum):
	NOAUTH = 0x00
	GSSAPI = 0x01
	PLAIN  = 0x02
	# IANA ASSIGNED X'03' to X'7F'
	# RESERVED FOR PRIVATE METHODS X'80' to X'FE'

	NOTACCEPTABLE = 0xFF


class SOCKS5Command(enum.Enum):
	CONNECT = 0x01
	BIND = 0x02
	UDP_ASSOCIATE = 0x03


class SOCKS5AddressType(enum.Enum):
	IP_V4 = 0x01
	DOMAINNAME = 0x03
	IP_V6 = 0x04


class SOCKS5ReplyType(enum.Enum):
	SUCCEEDED = 0X00 # o  X'00' succeeded
	FAILURE = 0x01 # o  X'01' general SOCKS server failure
	CONN_NOT_ALLOWED = 0x02#         o  X'02' connection not allowed by ruleset
	NETWORK_UNREACHABLE = 0x03 #o  X'03' Network unreachable
	HOST_UNREACHABLE = 0x04#o  X'04' Host unreachable
	CONN_REFUSED = 0x05 #o  X'05' Connection refused
	TTL_EXPIRED = 0x06 #o  X'06' TTL expired
	COMMAND_NOT_SUPPORTED = 0x07 #o  X'07' Command not supported
	ADDRESS_TYPE_NOT_SUPPORTED = 0x08 #o  X'08' Address type not supported
	#o  X'09' to X'FF' unassigned


class SOCKS5SocketParser:
	def __init__(self, protocol = socket.SOCK_STREAM):
		self.protocol = protocol

	def parse(self, soc, packet_type):
		return packet_type.from_bytes(self.read_soc(soc, packet_type.size))

	def read_soc(self, soc, size):
		data = b''
		while True:
			temp = soc.recv(4096)
			if temp == '':
				break
			data += temp
			if len(data) == size:
				break
		return data


class SOCKS5CommandParser:
	# the reason we need this class is: SOCKS5 protocol messages doesn't have a type field,
	# the messages are parsed in context of the session itself.
	def __init__(self, protocol = socket.SOCK_STREAM):
		self.protocol = protocol #not used atm

	def parse(self, buff, session):
		if session.current_state == SOCKS5ServerState.NEGOTIATION:
			return SOCKS5Nego.from_buffer(buff)
		
		if session.current_state == SOCKS5ServerState.NOT_AUTHENTICATED:
			if session.mutual_auth_type == SOCKS5Method.PLAIN:
				return SOCKS5PlainAuth.from_buffer(buff)
			else:
				raise Exception('Not implemented!')

		if session.current_state == SOCKS5ServerState.REQUEST:
			return SOCKS5Request.from_buffer(buff)

	@staticmethod
	async def from_streamreader(reader, session, timeout = None):
		if session.current_state == SOCKS5ServerState.NEGOTIATION:
			t = await asyncio.wait_for(SOCKS5Nego.from_streamreader(reader), timeout = timeout)
			return t
		
		if session.current_state == SOCKS5ServerState.NOT_AUTHENTICATED:
			if session.mutual_auth_type == SOCKS5Method.PLAIN:
				t = await asyncio.wait_for(SOCKS5PlainAuth.from_streamreader(reader), timeout = timeout)
				return t
			else:
				raise Exception('Not implemented!')

		if session.current_state == SOCKS5ServerState.REQUEST:
			t = await asyncio.wait_for(SOCKS5Request.from_streamreader(reader), timeout = timeout)
			return t


class SOCKS5AuthHandler:
	def __init__(self, authtype, creds = None):
		self.authtype  = authtype
		self.creds = creds

	def do_AUTH(self, msg):
		if self.authtype == SOCKS5Method.PLAIN:
			if not isinstance(msg, SOCKS5PlainAuth):
				raise Exception('Wrong message/auth type!')

			if self.creds is None:
				return True, SOCKS5PlainCredentials(msg.UNAME, msg.PASSWD)
			else:
				if msg.UNAME in self.creds:
					if msg.PASSWD == self.creds[msg.UNAME]:
						return True, SOCKS5PlainCredentials(msg.UNAME, msg.PASSWD)

				return False, SOCKS5PlainCredentials(msg.UNAME, msg.PASSWD)

		elif self.authtype == SOCKS5Method.GSSAPI:
			raise Exception('Not implemented! yet')
		
		else:
			raise Exception('Not implemented!')


class SOCKS5PlainCredentials:
	def __init__(self, username, password):
		self.username = username
		self.password = password


class SOCKS5PlainAuth:
	def __init__(self):
		self.VER = None
		self.ULEN = None
		self.UNAME = None
		self.PLEN = None
		self.PASSWD = None

	@staticmethod
	async def from_streamreader(reader, timeout = None):
		auth = SOCKS5PlainAuth()
		t = await read_or_exc(reader, 1, timeout = timeout)
		auth.VER = int.from_bytes(t, byteorder = 'big', signed = False)
		t = await read_or_exc(reader, 1, timeout = timeout)
		auth.ULEN = int.from_bytes(t, byteorder = 'big', signed = False)
		t = await read_or_exc(reader, auth.ULEN, timeout = timeout)
		auth.UNAME = t.decode()
		t = await read_or_exc(reader, 1, timeout = timeout)
		auth.PLEN = int.from_bytes(t, byteorder = 'big', signed = False)
		t = await read_or_exc(reader, auth.PLEN, timeout = timeout)
		auth.PASSWD = t.decode()

		return auth

	@staticmethod
	def from_bytes(bbuff):
		return SOCKS5PlainAuth.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		auth = SOCKS5PlainAuth()
		auth.VER = int.from_bytes(buff.read(1), byteorder = 'big', signed = False) 
		auth.ULEN = int.from_bytes(buff.read(1), byteorder = 'big', signed = False)
		auth.UNAME = buff.read(auth.ULEN).decode()
		auth.PLEN = int.from_bytes(buff.read(1), byteorder = 'big', signed = False)
		auth.PASSWD = buff.read(auth.PLEN).decode()

		return auth

	@staticmethod
	def construct(username, password):
		auth = SOCKS5PlainAuth()
		auth.VER    = 5
		auth.ULEN   = len(username)
		auth.UNAME  = username
		auth.PLEN   = len(password)
		auth.PASSWD = password

		return auth

	def to_bytes(self):
		t  = self.VER.to_bytes(1, byteorder = 'big', signed = False)
		t += self.ULEN.to_bytes(1, byteorder = 'big', signed = False)
		t += self.UNAME.encode()
		t += self.PLEN.to_bytes(1, byteorder = 'big', signed = False)
		t += self.PASSWD.encode()
		return t


class SOCKS5Nego:
	def __init__(self):
		self.VER = None
		self.NMETHODS = None
		self.METHODS = None

	@staticmethod
	async def from_streamreader(reader, timeout = None):
		nego = SOCKS5Nego()
		t = await read_or_exc(reader,1, timeout = timeout)
		nego.VER = int.from_bytes(t, byteorder = 'big', signed = False)
		t = await read_or_exc(reader,1, timeout = timeout)
		nego.NMETHODS = int.from_bytes(t, byteorder = 'big', signed = False)
		nego.METHODS = []
		for _ in range(nego.NMETHODS):
			t = await read_or_exc(reader,1, timeout = timeout)
			nego.METHODS.append(SOCKS5Method(int.from_bytes(t, byteorder = 'big', signed = False)))

		return nego

	@staticmethod
	def from_bytes(bbuff):
		return SOCKS5Nego.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		nego = SOCKS5Nego()
		nego.VER = int.from_bytes(buff.read(1), byteorder = 'big', signed = False) 
		nego.NMETHODS = int.from_bytes(buff.read(1), byteorder = 'big', signed = False) 
		nego.METHODS = []
		for _ in range(nego.NMETHODS):
			nego.METHODS.append(SOCKS5Method(int.from_bytes(buff.read(1), byteorder = 'big', signed = False)))
		return nego

	@staticmethod
	def construct(methods):
		if not isinstance(methods, list):
			methods = [methods]
		nego = SOCKS5Nego()
		nego.VER = 5
		nego.NMETHODS = len(methods)
		nego.METHODS = methods
		return nego

	def to_bytes(self):
		t  = self.VER.to_bytes(1, byteorder = 'big', signed = False)
		t += self.NMETHODS.to_bytes(1, byteorder = 'big', signed = False)
		for method in self.METHODS:
			t += method.value.to_bytes(1, byteorder = 'big', signed = False)
		return t

class SOCKS5NegoReply:
	def __init__(self):
		self.VER = None
		self.METHOD = None

	def __repr__(self):
		t  = '== SOCKS5NegoReply ==\r\n'
		t += 'VER: %s\r\n' % self.VER
		t += 'METHOD: %s\r\n' % self.METHOD
		return t

	@staticmethod
	def from_socket(soc):
		data = b''
		total_size = 2
		while True:
			temp = soc.recv(1024)
			if temp == b'':
				break
			data += temp
			if len(data) >= total_size:
				break
		return SOCKS5NegoReply.from_bytes(data)

	@staticmethod
	async def from_streamreader(reader, timeout = None):
		rep = SOCKS5NegoReply()
		t = await read_or_exc(reader,1, timeout = timeout)
		rep.VER = int.from_bytes(t, byteorder = 'big', signed = False)
		t = await read_or_exc(reader,1, timeout = timeout)
		rep.METHOD = SOCKS5Method(int.from_bytes(t, byteorder = 'big', signed = False))
		
		return rep

	@staticmethod
	def from_bytes(bbuff):
		return SOCKS5NegoReply.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		rep = SOCKS5NegoReply()
		rep.VER = int.from_bytes(buff.read(1), byteorder = 'big', signed = False)
		rep.METHOD = SOCKS5Method(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		return rep

	@staticmethod
	def construct(method):
		rep = SOCKS5NegoReply()
		rep.VER = 5
		rep.METHOD = method
		return rep

	@staticmethod
	def construct_auth(method, ver = 1):
		rep = SOCKS5NegoReply()
		rep.VER = ver
		rep.METHOD = method
		return rep


	def to_bytes(self):
		t  = self.VER.to_bytes(1, byteorder = 'big', signed = False)
		t += self.METHOD.value.to_bytes(1, byteorder = 'big', signed = False)
		return t


class SOCKS5Request:
	def __init__(self):
		self.VER = None
		self.CMD = None
		self.RSV = None
		self.ATYP = None
		self.DST_ADDR = None
		self.DST_PORT = None

	@staticmethod
	async def from_streamreader(reader, timeout = None):
		req = SOCKS5Request()
		t = await read_or_exc(reader,1, timeout = timeout)
		req.VER = int.from_bytes(t, byteorder = 'big', signed = False)
		t = await read_or_exc(reader,1, timeout = timeout)
		req.CMD = SOCKS5Command(int.from_bytes(t, byteorder = 'big', signed = False))
		t = await read_or_exc(reader,1, timeout = timeout)
		req.RSV = int.from_bytes(t, byteorder = 'big', signed = False)
		t = await read_or_exc(reader,1, timeout = timeout)
		req.ATYP = SOCKS5AddressType(int.from_bytes(t, byteorder = 'big', signed = False))
		if req.ATYP == SOCKS5AddressType.IP_V4:
			t = await read_or_exc(reader,4, timeout = timeout)
			req.DST_ADDR = ipaddress.IPv4Address(t)
		elif req.ATYP == SOCKS5AddressType.IP_V6:
			t = await read_or_exc(reader,16, timeout = timeout)
			req.DST_ADDR = ipaddress.IPv6Address(t)

		elif req.ATYP == SOCKS5AddressType.DOMAINNAME:
			t = await read_or_exc(reader,1, timeout = timeout)
			length = int.from_bytes(t, byteorder = 'big', signed = False)
			t = await read_or_exc(reader,length, timeout = timeout)
			req.DST_ADDR = t.decode()

		t = await read_or_exc(reader,2, timeout = timeout)
		req.DST_PORT = int.from_bytes(t, byteorder = 'big', signed = False)

		return req

	@staticmethod
	def from_bytes(bbuff):
		return SOCKS5Request.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		req = SOCKS5Request()
		req.VER = int.from_bytes(buff.read(1), byteorder = 'big', signed = False)
		req.CMD = SOCKS5Command(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		req.RSV = int.from_bytes(buff.read(1), byteorder = 'big', signed = False) 
		req.ATYP = SOCKS5AddressType(int.from_bytes(buff.read(1), byteorder = 'big', signed = False)) 
		if req.ATYP == SOCKS5AddressType.IP_V4:
			req.DST_ADDR = ipaddress.IPv4Address(buff.read(4))
		elif req.ATYP == SOCKS5AddressType.IP_V6:
			req.DST_ADDR = ipaddress.IPv6Address(buff.read(16))
		elif req.ATYP == SOCKS5AddressType.DOMAINNAME:
			length = int.from_bytes(buff.read(1), byteorder = 'big', signed = False) 
			req.DST_ADDR = buff.read(length).decode()

		req.DST_PORT = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		return req

	@staticmethod
	def construct(cmd, address, port):
		req = SOCKS5Request()
		req.VER = 5
		req.CMD = cmd
		req.RSV = 0
		if isinstance(address, ipaddress.IPv4Address):
			req.ATYP = SOCKS5AddressType.IP_V4
			req.DST_ADDR = address
		elif isinstance(address, ipaddress.IPv6Address):
			req.ATYP = SOCKS5AddressType.IP_V6
			req.DST_ADDR = address
		elif isinstance(address, str):
			req.ATYP = SOCKS5AddressType.DOMAINNAME
			req.DST_ADDR = address

		req.DST_PORT = port
		return req

	def to_bytes(self):
		t  = self.VER.to_bytes(1, byteorder = 'big', signed = False)
		t += self.CMD.value.to_bytes(1, byteorder = 'big', signed = False)
		t += self.RSV.to_bytes(1, byteorder = 'big', signed = False)
		t += self.ATYP.value.to_bytes(1, byteorder = 'big', signed = False)
		if self.ATYP == SOCKS5AddressType.DOMAINNAME:
			t += len(self.DST_ADDR).to_bytes(1, byteorder = 'big', signed = False)
			t += self.DST_ADDR.encode()
		else:	
			t += self.DST_ADDR.packed
		t += self.DST_PORT.to_bytes(2, byteorder = 'big', signed = False)
		return t


class SOCKS5Reply:
	def __init__(self):
		self.VER = None
		self.REP = None
		self.RSV = None
		self.ATYP = None
		self.BIND_ADDR= None
		self.BIND_PORT= None

	@staticmethod
	def from_socket(soc):
		data = b''
		total_size = 1024
		while True:
			temp = soc.recv(1024)
			if temp == b'':
				break
			data += temp

			if len(data) > 4:
				rt = SOCKS5AddressType(data[3])
				if rt == SOCKS5AddressType.IP_V4:
					total_size = 4 + 2 + 4
				if rt == SOCKS5AddressType.IP_V6:
					total_size = 4 + 2 + 16
				if rt == SOCKS5AddressType.DOMAINNAME:
					total_size = 4 + 2 + data[4]
			if len(data) >= total_size:
				break

		return SOCKS5Reply.from_bytes(data)

	@staticmethod
	async def from_streamreader(reader, timeout = None):
		rep = SOCKS5Reply()
		t = await read_or_exc(reader,1, timeout = timeout)
		rep.VER = int.from_bytes(t, byteorder = 'big', signed = False)
		t = await read_or_exc(reader,1, timeout = timeout)
		rep.REP = SOCKS5ReplyType(int.from_bytes(t, byteorder = 'big', signed = False))
		t = await read_or_exc(reader,1, timeout = timeout)
		rep.RSV = int.from_bytes(t, byteorder = 'big', signed = False)
		t = await read_or_exc(reader,1, timeout = timeout)
		rep.ATYP = SOCKS5AddressType(int.from_bytes(t, byteorder = 'big', signed = False))
		if rep.ATYP == SOCKS5AddressType.IP_V4:
			t = await read_or_exc(reader,4, timeout = timeout)
			rep.BIND_ADDR = ipaddress.IPv4Address(t)
		elif rep.ATYP == SOCKS5AddressType.IP_V6:
			t = await read_or_exc(reader,16, timeout = timeout)
			rep.BIND_ADDR = ipaddress.IPv6Address(t)
		elif rep.ATYP == SOCKS5AddressType.DOMAINNAME:
			t = await read_or_exc(reader,1, timeout = timeout)
			length = int.from_bytes(t, byteorder = 'big', signed = False)
			t = await read_or_exc(reader,length, timeout = timeout)
			rep.BIND_ADDR = t.decode()

		t = await read_or_exc(reader,2, timeout = timeout)
		rep.BIND_PORT = int.from_bytes(t, byteorder = 'big', signed = False)
		return rep

	@staticmethod
	def from_bytes(bbuff):
		return SOCKS5Reply.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		rep = SOCKS5Reply()
		rep.VER = int.from_bytes(buff.read(1), byteorder = 'big', signed = False) 
		rep.REP = SOCKS5ReplyType(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))
		rep.RSV = int.from_bytes(buff.read(1), byteorder = 'big', signed = False)
		rep.ATYP = SOCKS5AddressType(int.from_bytes(buff.read(1), byteorder = 'big', signed = False))

		if rep.ATYP == SOCKS5AddressType.IP_V4:
			rep.BIND_ADDR = ipaddress.IPv4Address(buff.read(4))
		elif rep.ATYP == SOCKS5AddressType.IP_V6:
			rep.BIND_ADDR = ipaddress.IPv6Address(buff.read(16))
		elif rep.ATYP == SOCKS5AddressType.DOMAINNAME:
			length = int.from_bytes(buff.read(1), byteorder = 'big', signed = False) 
			rep.BIND_ADDR = buff.read(length).decode()

		rep.BIND_PORT = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)

		return rep

	@staticmethod
	def construct(reply, address, port): 
		rep = SOCKS5Reply()
		rep.VER = 5
		rep.REP = reply
		rep.RSV = 0
		if isinstance(address, ipaddress.IPv4Address):
			rep.ATYP = SOCKS5AddressType.IP_V4
			rep.DST_ADDR = address
		elif isinstance(address, ipaddress.IPv6Address):
			rep.ATYP = SOCKS5AddressType.IP_V6
			rep.DST_ADDR = address
		elif isinstance(address, str):
			rep.ATYP = SOCKS5AddressType.DOMAINNAME
			rep.DST_ADDR = address

		rep.DST_PORT = port
		return rep

	def to_bytes(self):
		t  = self.VER.to_bytes(1, byteorder = 'big', signed = False)
		t += self.REP.value.to_bytes(1, byteorder = 'big', signed = False)
		t += self.RSV.to_bytes(1, byteorder = 'big', signed = False)
		t += self.ATYP.value.to_bytes(1, byteorder = 'big', signed = False)
		if self.ATYP == SOCKS5AddressType.DOMAINNAME:
			t += len(self.DST_ADDR).to_bytes(1, byteorder = 'big', signed = False)
			t += self.DST_ADDR.encode()
		else:	
			t += self.DST_ADDR.packed
		t += self.DST_PORT.to_bytes(2, byteorder = 'big', signed = False)
		return t

	def __repr__(self):
		t  = '== SOCKS5Reply ==\r\n'
		t += 'REP: %s\r\n' % repr(self.REP)
		t += 'ATYP: %s\r\n' % repr(self.ATYP)
		t += 'BIND_ADDR: %s\r\n' % repr(self.BIND_ADDR)
		t += 'BIND_PORT: %s\r\n' % repr(self.BIND_PORT)

		return t


class SOCKS5UDP:
	def __init__(self):
		self.RSV = None
		self.FRAG = None
		self.ATYP = None
		self.DST_ADDR = None
		self.DST_PORT = None
		self.DATA = None

	@staticmethod
	async def from_streamreader(reader, timeout = None):
		rep = SOCKS5UDP()
		t = await read_or_exc(reader,2, timeout = timeout)
		rep.RSV = int.from_bytes(t, byteorder = 'big', signed = False)
		t = await read_or_exc(reader,1, timeout = timeout)
		rep.FRAG = SOCKS5ReplyType(int.from_bytes(t, byteorder = 'big', signed = False))
		t = await read_or_exc(reader,1, timeout = timeout)
		rep.ATYP = SOCKS5AddressType(int.from_bytes(t, byteorder = 'big', signed = False))
		if rep.ATYP == SOCKS5AddressType.IP_V4:
			t = await read_or_exc(reader,4, timeout = timeout)
			rep.DST_ADDR = ipaddress.IPv4Address(t)
		elif rep.ATYP == SOCKS5AddressType.IP_V6:
			t = await read_or_exc(reader,16, timeout = timeout)
			rep.DST_ADDR = ipaddress.IPv6Address(t)

		elif rep.ATYP == SOCKS5AddressType.DOMAINNAME:
			t = await read_or_exc(reader,1, timeout = timeout)
			length = int.from_bytes(t, byteorder = 'big', signed = False)
			t = await read_or_exc(reader,length, timeout = timeout)
			rep.DST_ADDR = t.decode()

		t = await read_or_exc(reader,2, timeout = timeout)
		rep.DST_PORT = int.from_bytes(t, byteorder = 'big', signed = False)
		return rep

	@staticmethod
	def from_bytes(bbuff):
		return SOCKS5UDP.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		rep = SOCKS5UDP()
		rep.RSV = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		rep.FRAG = int.from_bytes(buff.read(1), byteorder = 'big', signed = False)
		rep.ATYP = SOCKS5AddressType(buff.read(1), byteorder = 'big', signed = False)
		if rep.ATYP == SOCKS5AddressType.IP_V4:
			rep.DST_ADDR = ipaddress.IPv4Address(buff.read(4))
		elif rep.ATYP == SOCKS5AddressType.IP_V6:
			rep.DST_ADDR = ipaddress.IPv6Address(buff.read(16))
		elif rep.ATYP == SOCKS5AddressType.DOMAINNAME:
			length = int.from_bytes(buff.read(1), byteorder = 'big', signed = False) 
			rep.DST_ADDR = buff.read(length).decode()

		rep.DST_PORT = int.from_bytes(buff.read(2), byteorder = 'big', signed = False)
		#be careful, not data length is defined in the RFC!!
		rep.DATA = buff.read()

	@staticmethod
	def construct(address, port, data, frag = 0):
		req = SOCKS5Request()
		req.RSV = 0
		req.FRAG = frag
		if isinstance(address, ipaddress.IPv4Address):
			req.ATYP = SOCKS5AddressType.IP_V4
			req.DST_ADDR = address
		elif isinstance(address, ipaddress.IPv6Address):
			req.ATYP = SOCKS5AddressType.IP_V6
			req.DST_ADDR = address
		elif isinstance(address, str):
			req.ATYP = SOCKS5AddressType.DOMAINNAME
			req.DST_ADDR = address

		req.DST_PORT = port
		req.DATA = data
		return req

	def to_bytes(self):
		t  = self.RSV.to_bytes(2, byteorder = 'big', signed = False)
		t += self.FRAG.value.to_bytes(1, byteorder = 'big', signed = False)
		t += self.ATYP.value.to_bytes(1, byteorder = 'big', signed = False)
		if self.ATYP == SOCKS5AddressType.DOMAINNAME:
			t += len(self.DST_ADDR).to_bytes(1, byteorder = 'big', signed = False)
			t += self.DST_ADDR.encode()
		else:	
			t += self.DST_ADDR.packed
		t += self.DST_PORT.to_bytes(2, byteorder = 'big', signed = False)
		t += self.DATA
		return t


from aiosmb.protocol.smb2.commands.negotiate import NEGOTIATE_REQ, NEGOTIATE_REPLY,NegotiateSecurityMode, NegotiateCapabilities, NegotiateDialects
from aiosmb.protocol.smb2.commands.sessionsetup import SESSION_SETUP_REQ, SESSION_SETUP_REPLY
from aiosmb.protocol.smb2.commands.tree_connect import TREE_CONNECT_REQ, TREE_CONNECT_REPLY, TreeConnectFlag, TreeCapabilities, ShareFlags
from aiosmb.protocol.smb2.commands.create import CREATE_REQ, CREATE_REPLY, OplockLevel, ImpersonationLevel, OplockLevel, ImpersonationLevel, ShareAccess, CreateDisposition, CreateOptions
from aiosmb.protocol.smb2.commands.read import READ_REQ, READ_REPLY, Channel, ReadFlag
from aiosmb.protocol.smb2.commands.query_info import QUERY_INFO_REPLY, QUERY_INFO_REQ, EaInformation, SecurityInfo, QueryInfoType
from aiosmb.protocol.smb2.commands.query_directory import QUERY_DIRECTORY_REPLY, QUERY_DIRECTORY_REQ, QueryDirectoryFlag
from aiosmb.protocol.smb2.commands.tree_disconnect import TREE_DISCONNECT_REQ, TREE_DISCONNECT_REPLY
from aiosmb.protocol.smb2.commands.close import CLOSE_REQ, CLOSE_REPLY, CloseFlag
from aiosmb.protocol.smb2.commands.flush import FLUSH_REQ, FLUSH_REPLY
from aiosmb.protocol.smb2.commands.echo import ECHO_REQ, ECHO_REPLY
from aiosmb.protocol.smb2.commands.cancel import CANCEL_REQ
from aiosmb.protocol.smb2.commands.logoff import LOGOFF_REQ, LOGOFF_REPLY
from aiosmb.protocol.smb2.commands.error import ERROR_REPLY
from aiosmb.protocol.smb2.commands.write import WRITE_REPLY, WRITE_REQ



__all__ = ['NEGOTIATE_REQ', 'NEGOTIATE_REPLY', 'SESSION_SETUP_REQ','SESSION_SETUP_REPLY','NegotiateSecurityMode', 
			'NegotiateCapabilities', 'NegotiateDialects', 'TREE_CONNECT_REQ', 'TREE_CONNECT_REPLY','CREATE_REQ', 
			'CREATE_REPLY','OplockLevel', 'ImpersonationLevel','TreeConnectFlag', 'TreeCapabilities', 'ShareFlags',
			'ShareAccess', 'CreateDisposition', 'CreateOptions', 'READ_REQ', 'READ_REPLY', 'Channel', 
			'ReadFlag', 'QUERY_INFO_REPLY', 'QUERY_INFO_REQ', 'EaInformation', 'SecurityInfo', 'QueryInfoType',
			'QUERY_DIRECTORY_REPLY', 'QUERY_DIRECTORY_REQ', 'QueryDirectoryFlag','TREE_DISCONNECT_REQ','TREE_DISCONNECT_REPLY',
			'CLOSE_REQ','CLOSE_REPLY','FLUSH_REQ','FLUSH_REPLY','ECHO_REQ','ECHO_REPLY','CANCEL_REQ','LOGOFF_REQ',
			'LOGOFF_REPLY','ERROR_REPLY', 'CloseFlag', 'WRITE_REPLY', 'WRITE_REQ']
			
			
			



import asyncio
import traceback
import ntpath
import fnmatch

import shlex
import tqdm
from aiocmd import aiocmd
from aiosmb.examples.smbpathcompleter import SMBPathCompleter

from aiosmb import logger
from aiosmb._version import __banner__
from aiosmb.commons.connection.url import SMBConnectionURL
from aiosmb.commons.interfaces.machine import SMBMachine
from aiosmb.commons.utils.decorators import rr, rr_gen, red, red_gen, ef_gen
from aiosmb.commons.exceptions import SMBException, SMBMachineException
from aiosmb.dcerpc.v5.rpcrt import DCERPCException

from asysocks import logger as sockslogger


def req_traceback(funct):
	async def wrapper(*args, **kwargs):
		this = args[0]
		try:
			x = await funct(*args, **kwargs)
			return x
		except Exception as e:
			traceback.print_exc()
			raise e
	return wrapper

class SMBClient(aiocmd.PromptToolkitCmd):
	def __init__(self, url = None):
		aiocmd.PromptToolkitCmd.__init__(self, ignore_sigint=False) #Setting this to false, since True doesnt work on windows...
		self.conn_url = None
		if url is not None:
			self.conn_url = SMBConnectionURL(url)
		self.connection = None
		self.machine = None

		self.shares = {} #name -> share
		self.__current_share = None
		self.__current_directory = None

	async def do_login(self, url = None):
		"""Connects to the remote machine"""
		try:
			if self.conn_url is None and url is None:
				print('No url was set, cant do logon')
			if url is not None:
				self.conn_url = SMBConnectionURL(url)
			
			self.connection  = self.conn_url.get_connection()
			
			logger.debug(self.conn_url.get_credential())
			logger.debug(self.conn_url.get_target())

			await self.connection.login()
			self.machine = SMBMachine(self.connection)
		except Exception as e:
			traceback.print_exc()
		else:
			print('Login success')

	async def do_shares(self, show = True):
		"""Lists available shares"""
		try:
			async for share, err in ef_gen(self.machine.list_shares()):
				if err is not None:
					raise err
				self.shares[share.name] = share
				if show is True:
					print(share.name)
				
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_sessions(self):
		"""Lists sessions of connected users"""
		try:
			async for sess, err in ef_gen(self.machine.list_sessions()):
				if err is not None:
					raise err
				print("%s : %s" % (sess.username, sess.ip_addr))
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_domains(self):
		"""Lists domain"""
		try:
			async for domain, err in self.machine.list_domains():
				if err is not None:
					raise err
				print(domain)
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_localgroups(self):
		"""Lists local groups"""
		try:
			async for name, sid, err in self.machine.list_localgroups():
				if err is not None:
					raise err
				print("%s : %s" % (name, sid))
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()
	
	async def do_domaingroups(self, domain_name):
		"""Lists groups in a domain"""
		try:
			async for name, sid, err in self.machine.list_groups(domain_name):
				if err is not None:
					raise err
				print("%s : %s" % (name, sid))
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()
	
	async def do_groupmembers(self, domain_name, group_name):
		"""Lists members of an arbitrary group"""
		try:
			async for domain, username, sid, err in self.machine.list_group_members(domain_name, group_name):
				if err is not None:
					raise err
				print("%s\\%s : %s" % (domain, username, sid))
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_localgroupmembers(self, group_name):
		"""Lists members of a local group"""
		try:
			async for domain, username, sid, err in self.machine.list_group_members('Builtin', group_name):
				if err is not None:
					raise err
				print("%s\\%s : %s" % (domain, username, sid))
			
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_use(self, share_name):
		"""selects share to be used"""
		try:
			if len(self.shares) == 0:
				await self.do_shares(show = False)

			if share_name not in self.shares:
				if share_name.upper() not in self.shares:
					print('Error! Uknown share name %s' % share_name)
					return
				share_name = share_name.upper()

			self.__current_share = self.shares[share_name]
			await self.__current_share.connect(self.connection)
			self.__current_directory = self.__current_share.subdirs[''] #this is the entry directory
			self.prompt = '[%s] $' % self.__current_directory.unc_path
			await self.do_ls(False)

		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()
			
	
	async def do_ls(self, show = True):
		try:
			if self.__current_share is None:
				print('No share selected!')
				return
			if self.__current_directory is None:
				print('No directory selected!')
				return
			
			#print(self.__current_directory)
			async for entry in self.machine.list_directory(self.__current_directory):
				if show == True:
					print(entry)
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_cd(self, directory_name):
		try:
			if self.__current_share is None:
				print('No share selected!')
				return
			if self.__current_directory is None:
				print('No directory selected!')
				return
			
			if directory_name not in self.__current_directory.subdirs:
				if directory_name == '..':
					self.__current_directory = self.__current_directory.parent_dir
					self.prompt = '[%s] $' % (self.__current_directory.unc_path)
					return
				else:
					print('The directory "%s" is not in parent directory "%s"' % (directory_name, self.__current_directory.fullpath))
			
			else:
				self.__current_directory = self.__current_directory.subdirs[directory_name]
				self.prompt = '[%s] $' % (self.__current_directory.unc_path)
				await self.do_ls(False)
			
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	def get_current_dirs(self):
		if self.__current_directory is None:
			return []
		return list(self.__current_directory.subdirs.keys())

	def get_current_files(self):
		if self.__current_directory is None:
			return []
		return list(self.__current_directory.files.keys())

	async def do_sid(self, file_name):
		if file_name not in self.__current_directory.files:
			print('file not in current directory!')
			return
		file_obj = self.__current_directory.files[file_name]
		sid = await file_obj.get_security_descriptor(self.connection)
		print(str(sid))

	async def do_dirsid(self):
		sid = await self.__current_directory.get_security_descriptor(self.connection)
		print(str(sid))

	def _cd_completions(self):
		return SMBPathCompleter(get_current_dirs = self.get_current_dirs)

	def _get_completions(self):
		return SMBPathCompleter(get_current_dirs = self.get_current_files)

	async def do_services(self):
		"""Lists remote services"""
		try:
			async for service, err in self.machine.list_services():
				if err is not None:
					raise err
				print(service)
			
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_serviceen(self, service_name):
		"""Enables a remote service"""
		try:
			res, err = await self.machine.enable_service(service_name)
			if err is not None:
				raise err
			print(res)
		
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_servicecreate(self, service_name, command, display_name = None):
		"""Creates a remote service"""
		try:
			res, err = await self.machine.create_service(service_name, command, display_name)
			if err is not None:
				raise err
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_servicedeploy(self, path_to_exec, remote_path):
		"""Deploys a binary file from the local system as a service on the remote system"""
		#servicedeploy /home/devel/Desktop/cmd.exe /shared/a.exe
		try:
			basename = ntpath.basename(remote_path)
			remote_path = '\\\\%s\\%s\\%s\\%s' % (self.connection.target.get_hostname_or_ip(), self.__current_share.name, self.__current_directory.fullpath , basename)
			await rr(self.machine.deploy_service(path_to_exec, remote_path = remote_path))
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_put(self, file_name):
		"""Uploads a file to the remote share"""
		try:
			basename = ntpath.basename(file_name)
			dst = '\\%s\\%s\\%s' % (self.__current_share.name, self.__current_directory.fullpath , basename)
			print(basename)
			print(dst)
			await self.machine.put_file(file_name, dst)
			
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_del(self, file_name):
		"""Removes a file from the remote share"""
		try:
			basename = ntpath.basename(file_name)
			dst = '\\%s\\%s\\%s' % (self.__current_share.name, self.__current_directory.fullpath , basename)
			print(dst)
			await self.machine.del_file(dst)
			
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_regsave(self, hive_name, file_path):
		"""Saves a registry hive to a file on remote share"""
		try:
			await rr(self.machine.save_registry_hive(hive_name, file_path))
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_get(self, file_name):
		"""Download a file from the remote share to the current folder"""
		try:
			matched = []
			if file_name not in self.__current_directory.files:
				
				for fn in fnmatch.filter(list(self.__current_directory.files.keys()), file_name):
					matched.append(fn)
				if len(matched) == 0:
					print('File with name %s is not present in the directory %s' % (file_name, self.__current_directory.name))
					return
			else:
				matched.append(file_name)
			
			for file_name in matched:
				file_obj = self.__current_directory.files[file_name]
				with tqdm.tqdm(desc = 'Downloading %s' % file_name, total=file_obj.size, unit='B', unit_scale=True, unit_divisor=1024) as pbar:
					with open(file_name, 'wb') as outfile:
						async for data in self.machine.get_file_data(file_obj):
							if data is None:
								break
							outfile.write(data)
							pbar.update(len(data))
	
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	
	async def do_mkdir(self, directory_name):
		"""Creates a directory on the remote share"""
		try:
			await self.machine.create_subdirectory(directory_name, self.__current_directory)

		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_dcsync(self):
		"""It's a suprse tool that will help us later"""
		try:
			async for secret, err in self.machine.dcsync():
				if err is not None:
					raise err
				print(str(secret))
		
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_users(self, domain = None):
		"""List users in domain"""
		try:
			async for username, user_sid, err in self.machine.list_domain_users(domain):
				if err is not None:
					print(str(err))
				print('%s %s' % (username, user_sid))
		
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

	async def do_printerbug(self, attacker_ip):
		"""Printerbug"""
		try:
			res, err = await self.machine.printerbug(attacker_ip)
			if err is not None:
				print(str(err))
			print(res)
		
		except SMBException as e:
			logger.debug(traceback.format_exc())
			print(e.pprint())
		except SMBMachineException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except DCERPCException as e:
			logger.debug(traceback.format_exc())
			print(str(e))
		except Exception as e:
			traceback.print_exc()

			

async def amain(args):
	client = SMBClient(args.smb_url)
	if len(args.commands) == 0:
		if args.no_interactive is True:
			print('Not starting interactive!')
			return
		await client.run()
	else:
		for command in args.commands:
			cmd = shlex.split(command)
			#print(cmd)
			await client._run_single_command(cmd[0], cmd[1:])

def main():
	import argparse
	import platform
	
	parser = argparse.ArgumentParser(description='Interactive SMB client')
	parser.add_argument('-v', '--verbose', action='count', default=0)
	parser.add_argument('-n', '--no-interactive', action='store_true')
	parser.add_argument('smb_url', help = 'Connection string that describes the authentication and target. Example: smb+ntlm-password://TEST\\Administrator:password@10.10.10.2')
	parser.add_argument('commands', nargs='*')
	
	args = parser.parse_args()
	print(__banner__)

	if args.verbose > 2:
		print('setting deepdebug')
		logger.setLevel(1) #enabling deep debug
		sockslogger.setLevel(1)

	asyncio.run(amain(args))

if __name__ == '__main__':
	main()
	
	

	


import enum

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-fscc/4718fc40-e539-4014-8e33-b675af74e3e1
class FileInfoClass(enum.Enum):
	NONE = 0
	FileAccessInformation = 8 #Query
	FileAlignmentInformation = 17 #Query
	FileAllInformation = 18 #Query
	FileAllocationInformation = 19 #Set
	FileAlternateNameInformation = 21 #Query
	FileAttributeTagInformation = 35 #Query
	FileBasicInformation =4 #Query, Set
	FileBothDirectoryInformation = 3 #Query
	FileCompressionInformation = 28 #Query
	FileDirectoryInformation = 1 #Query
	FileDispositionInformation = 13 # Set
	FileEaInformation = 7 #Query
	FileEndOfFileInformation = 20 #Set
	FileFullDirectoryInformation = 2#Query
	FileFullEaInformation = 15 #Query, Set
	FileHardLinkInformation = 46 #LOCAL
	FileIdBothDirectoryInformation = 37 #Query
	FileIdFullDirectoryInformation = 38 #Query
	FileIdGlobalTxDirectoryInformation = 50 #LOCAL
	FileInternalInformation = 6 #Query
	FileLinkInformation = 11 #Set
	FileMailslotQueryInformation = 26 #LOCAL
	FileMailslotSetInformation = 27 #LOCAL
	FileModeInformation = 16 #Query, Set<78>
	FileMoveClusterInformation = 31 #<79>
	FileNameInformation = 9 #LOCAL
	FileNamesInformation = 12 #Query
	FileNetworkOpenInformation = 34 #Query
	FileNormalizedNameInformation = 48 #Query<80>
	FileObjectIdInformation = 29 #LOCAL
	FilePipeInformation = 23 #Query, Set
	FilePipeLocalInformation = 24 #Query
	FilePipeRemoteInformation = 25 #Query
	FilePositionInformation = 14 #Query, Set
	FileQuotaInformation = 32 #Query, Set<81>
	FileRenameInformation = 10 #Set
	FileReparsePointInformation = 33 #LOCAL
	FileSfioReserveInformation = 44 #LOCAL 
	FileSfioVolumeInformation = 45 #<82>
	FileShortNameInformation = 40 #Set
	FileStandardInformation = 5#Query
	FileStandardLinkInformation = 54 #LOCAL
	FileStreamInformation = 22 #Query
	FileTrackingInformation = 36 #LOCAL
	FileValidDataLengthInformation = 39 #Set


import os
import io

from aiosmb.authentication.ntlm.structures.fields import Fields
from aiosmb.authentication.ntlm.structures.negotiate_flags import NegotiateFlags
from aiosmb.authentication.ntlm.structures.version import Version
from aiosmb.authentication.ntlm.structures.avpair import AVPairs

from aiosmb.authentication.ntlm.templates.server import NTLMServerTemplates

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-nlmp/801a4681-8809-4be9-ab0d-61dcfe762786
class NTLMChallenge:
	def __init__(self):
		self.Signature         = b'NTLMSSP\x00'
		self.MessageType       = 2
		self.TargetNameFields  = None
		self.NegotiateFlags    = None
		self.ServerChallenge   = None
		self.Reserved          = b'\x00'*8
		self.TargetInfoFields  = None
		self.Version           = None
		self.Payload           = None

		self.TargetName        = None
		self.TargetInfo        = None
		
		
	@staticmethod
	def from_bytes(bbuff):
		return NTLMChallenge.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		t = NTLMChallenge()
		t.Signature         = buff.read(8)
		t.MessageType       = int.from_bytes(buff.read(4), byteorder = 'little', signed = False)
		t.TargetNameFields  = Fields.from_buffer(buff)
		t.NegotiateFlags    = NegotiateFlags(int.from_bytes(buff.read(4), byteorder = 'little', signed = False))
		t.ServerChallenge   = buff.read(8)
		t.Reserved          = buff.read(8)
		t.TargetInfoFields  = Fields.from_buffer(buff)
		
		if t.NegotiateFlags & NegotiateFlags.NEGOTIATE_VERSION: 
			t.Version = Version.from_buffer(buff)
			
		currPos = buff.tell()
		t.Payload = buff.read()
			
		if t.TargetNameFields.length != 0:
			buff.seek(t.TargetNameFields.offset, io.SEEK_SET)
			raw_data = buff.read(t.TargetNameFields.length)
			try:
				t.TargetName = raw_data.decode('utf-16le')
			except UnicodeDecodeError:
				# yet another cool bug. 
				t.TargetName = raw_data.decode('utf-8')
				
		if t.TargetInfoFields.length != 0:
			buff.seek(t.TargetInfoFields.offset, io.SEEK_SET)
			raw_data = buff.read(t.TargetInfoFields.length)
			t.TargetInfo = AVPairs.from_bytes(raw_data)
			
		
		
		return t

	@staticmethod
	def construct_from_template(templateName, challenge = os.urandom(8), ess = True):
		version    = NTLMServerTemplates[templateName]['version']
		challenge  = challenge
		targetName = NTLMServerTemplates[templateName]['targetname']
		targetInfo = NTLMServerTemplates[templateName]['targetinfo']
		targetInfo = NTLMServerTemplates[templateName]['targetinfo']
		flags      = NTLMServerTemplates[templateName]['flags']
		if ess:
			flags |= NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY
		else:
			flags &= ~NegotiateFlags.NEGOTIATE_EXTENDED_SESSIONSECURITY

		return NTLMChallenge.construct(challenge=challenge, targetName = targetName, targetInfo = targetInfo, version = version, flags= flags)
	
	
	# TODO: needs some clearning up (like re-calculating flags when needed)
	@staticmethod
	def construct(challenge = os.urandom(8), targetName = None, targetInfo = None, version = None, flags = None):
		pos = 48
		if version:
			pos += 8
		t = NTLMChallenge()
		t.NegotiateFlags    = flags
		t.Version           = version
		t.ServerChallenge   = challenge
		t.TargetName        = targetName
		t.TargetInfo        = targetInfo

		t.TargetNameFields = Fields(len(t.TargetName.encode('utf-16le')),pos) 
		t.TargetInfoFields = Fields(len(t.TargetInfo.to_bytes()), pos + len(t.TargetName.encode('utf-16le')))

		t.Payload = t.TargetName.encode('utf-16le')
		t.Payload += t.TargetInfo.to_bytes()

		return t

	def to_bytes(self):
		tn = self.TargetName.encode('utf-16le')
		ti = self.TargetInfo.to_bytes()

		buff  = self.Signature
		buff += self.MessageType.to_bytes(4, byteorder = 'little', signed = False)
		buff += self.TargetNameFields.to_bytes()
		buff += self.NegotiateFlags.to_bytes(4, byteorder = 'little', signed = False)
		buff += self.ServerChallenge
		buff += self.Reserved
		buff += self.TargetInfoFields.to_bytes()
		if self.Version:
			buff += self.Version.to_bytes()
		buff += self.Payload

		return buff

	def __repr__(self):
		t  = '== NTLMChallenge ==\r\n'
		t += 'Signature      : %s\r\n' % repr(self.Signature)
		t += 'MessageType    : %s\r\n' % repr(self.MessageType)
		t += 'ServerChallenge: %s\r\n' % repr(self.ServerChallenge)
		t += 'TargetName     : %s\r\n' % repr(self.TargetName)
		t += 'TargetInfo     : %s\r\n' % repr(self.TargetInfo)
		return t

	def toBase64(self):
		return base64.b64encode(self.to_bytes()).decode('ascii')


def test():
	test_reconstrut()
	test_construct()
	test_template()
	
def test_reconstrut(data = None):
	print('=== reconstruct===')
	if not data:
		challenge_test_data = bytes.fromhex('4e544c4d53535000020000000800080038000000158289e2a7314a557bdb11bf000000000000000072007200400000000a0063450000000f540045005300540002000800540045005300540001001200570049004e003200300031003900410044000400120074006500730074002e0063006f007200700003002600570049004e003200300031003900410044002e0074006500730074002e0063006f007200700007000800aec600bfc5fdd40100000000')
	else:
		challenge_test_data = data
	challenge = NTLMChallenge.from_bytes(challenge_test_data)
	print(repr(challenge))
	challenge_test_data_verify = challenge.to_bytes()
	print('====== reconstructed ====')
	print(hexdump(challenge_test_data_verify))
	print('====== original ====')
	print(hexdump(challenge_test_data))
	assert challenge_test_data == challenge_test_data_verify
	
def test_template():
	
	challenge = NTLMChallenge.construct_from_template('Windows2003')
	test_reconstrut(challenge.to_bytes())
	
def test_construct():
	pass
	
if __name__ == '__main__':
	from aiosmb.utils.hexdump import hexdump
	
	test()



# THIS CODE IS UNDER MIT LICENSE AND WAS TAKEN FROM THE FOLLOWING URL
# https://github.com/eerykitty/CVE-2020-0796-PoC/blob/master/smbprotocol/lznt1.py

# MIT License
# 
# Copyright (c) 2017 Jordan Borean, Red Hat
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import struct
import sys
import copy

def _decompress_chunk(chunk):
    size = len(chunk)
    out = bytearray()
    pow2 = 0x10
    while chunk:
        flags = chunk[0]
        chunk = chunk[1:]
        for i in range(8):
            out_start = len(out)
            if not (flags >> i & 1):
                out += chunk[0].to_bytes(length=1, byteorder='little')
                chunk = chunk[1:]
            else:
                flag = struct.unpack('<H', chunk[:2])[0]
                pos = len(out) - 1
                l_mask = 0xFFF
                o_shift = 12
                while pos >= 0x10:
                    l_mask >>= 1
                    o_shift -= 1
                    pos >>= 1

                length = (flag & l_mask) + 3
                offset = (flag >> o_shift) + 1

                if length >= offset:
                    tmp = out[-offset:] * int(0xFFF / len(out[-offset:]) + 1)
                    out += tmp[:length]
                else:
                    out += out[-offset:-offset+length]
                chunk = chunk[2:]
            if len(chunk) == 0:
                break
    return out

def decompress(buf, length_check=True):
    out = bytearray()
    while buf:
        header = struct.unpack('<H', bytes(buf[:2]))[0]
        length = (header & 0xFFF) + 1
        if length_check and length > len(buf[2:]):
            raise ValueError('invalid chunk length')
        else:
            chunk = buf[2:2+length]
            if header & 0x8000:
                out += _decompress_chunk(chunk)
            else:
                out += chunk
        buf = buf[2+length:]
    return out

def _find(src, target, max_len):
    result_offset = 0
    result_length = 0
    for i in range(1, max_len):
        offset = src.rfind(target[:i])
        if offset == -1:
            break
        tmp_offset = len(src) - offset
        tmp_length = i
        if tmp_offset == tmp_length:
            tmp = src[offset:] * int(0xFFF / len(src[offset:]) + 1)
            for j in range(i, max_len+1):
                offset = tmp.rfind(target[:j])
                if offset == -1:
                    break
                tmp_length = j
        if tmp_length > result_length:
            result_offset = tmp_offset
            result_length = tmp_length

    if result_length < 3:
        return 0, 0
    return result_offset, result_length

def _compress_chunk(chunk):
    blob = copy.copy(chunk)
    out = b""
    pow2 = 0x10
    l_mask3 = 0x1002
    o_shift = 12
    while len(blob) > 0:
        bits = 0
        tmp = b""
        for i in range(8):
            bits >>= 1
            while pow2 < (len(chunk) - len(blob)):
                pow2 <<= 1
                l_mask3 = (l_mask3 >> 1) + 1
                o_shift -= 1
            if len(blob) < l_mask3:
                max_len = len(blob)
            else:
                max_len = l_mask3

            offset, length = _find(chunk[:len(chunk) - len(blob)], blob, max_len)

            # try to find more compressed pattern
            offset2, length2 = _find(chunk[:len(chunk) - len(blob)+1], blob[1:], max_len)
            if length < length2:
                length = 0

            if length > 0:
                symbol = ((offset-1) << o_shift) | (length - 3)
                tmp += struct.pack('<H', symbol)
                bits |= 0x80 # set the highest bit
                blob = blob[length:]
            else:
                tmp += bytes([blob[0]])
                blob = blob[1:]
            if len(blob) == 0:
                break

        out += struct.pack('B', bits >> (7 - i))
        out += tmp

    return out

def compress(buf, chunk_size=0x1000):
    out = b""
    while buf:
        chunk = buf[:chunk_size]
        compressed = _compress_chunk(chunk)
        if len(compressed) < len(chunk): # chunk is compressed
            flags = 0xB000
            header = struct.pack('<H' , flags|(len(compressed)-1))
            out += header + compressed
        else:
            flags = 0x3000
            header = struct.pack('<H' , flags|(len(chunk)-1))
            out += header + chunk
        buf = buf[chunk_size:]

    return out




import enum

class SMBSecurityMode(enum.IntFlag):
	NEGOTIATE_USER_SECURITY = 0x01
	NEGOTIATE_ENCRYPT_PASSWORDS = 0x02
	NEGOTIATE_SECURITY_SIGNATURES_ENABLED = 0x04
	NEGOTIATE_SECURITY_SIGNATURES_REQUIRED = 0x08
	#others are Reserved


class SMBCapabilities(enum.IntFlag):
	CAP_RAW_MODE         = 0x00000001
	CAP_MPX_MODE         = 0x00000002
	CAP_UNICODE          = 0x00000004
	CAP_LARGE_FILES      = 0x00000008
	CAP_NT_SMBS          = 0x00000010
	CAP_RPC_REMOTE_APIS  = 0x00000020
	CAP_STATUS32         = 0x00000040
	CAP_LEVEL_II_OPLOCKS = 0x00000080
	CAP_LOCK_AND_READ    = 0x00000100
	CAP_NT_FIND          = 0x00000200
	CAP_BULK_TRANSFER    = 0x00000400
	CAP_COMPRESSED_DATA  = 0x00000800
	CAP_DFS              = 0x00001000
	CAP_QUADWORD_ALIGNED = 0x00002000
	CAP_LARGE_READX      = 0x00004000
	CAP_NT_EXTENDED_SECURITY = 0x80000000


import io
import enum

from aiosmb.wintypes.fscc.FileAttributes import *

class CloseFlag(enum.IntFlag):
	NONE = 0
	SMB2_CLOSE_FLAG_POSTQUERY_ATTRIB = 0x0001 #If set, the server MUST set the attribute fields in the response, as specified in section 2.2.16, to valid values. If not set, the client MUST NOT use the values that are returned in the response.

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/f84053b0-bcb2-4f85-9717-536dae2b02bd
class CLOSE_REQ:
	def __init__(self):
		self.StructureSize = 24
		self.Flags = 0
		self.Reserved = 0
		self.FileId = None
		
	def to_bytes(self):
		t  = self.StructureSize.to_bytes(2, byteorder='little', signed = False)
		t += self.Flags.value.to_bytes(2, byteorder='little', signed = False)
		t += self.Reserved.to_bytes(4, byteorder='little', signed = False)
		t += self.FileId.to_bytes(16, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff):
		return CLOSE_REQ.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = CLOSE_REQ()
		msg.StructureSize   = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 24
		msg.Flags  = CloseFlag(int.from_bytes(buff.read(2), byteorder='little'))
		msg.Reserved  = int.from_bytes(buff.read(4), byteorder='little')
		msg.FileId  = int.from_bytes(buff.read(16), byteorder='little')
		return msg

	def __repr__(self):
		t = '==== SMB2 CLOSE REQ ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'Flags: %s\r\n' % self.Flags
		t += 'FileId: %s\r\n' % self.FileId
		return t
		
		
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/c0c15c57-3f3e-452b-b51c-9cc650a13f7b
class CLOSE_REPLY:
	def __init__(self):
		self.StructureSize = 60
		self.Flags = None
		self.Reserved = 0
		self.CreationTime = None
		self.LastAccessTime = None
		self.LastWriteTime = None
		self.ChangeTime = None
		self.AllocationSize = None
		self.EndofFile = None
		self.FileAttributes = None
		
	def to_bytes(self):
		t  = self.StructureSize.to_bytes(2, byteorder='little', signed = False)
		t += self.Flags.value.to_bytes(2, byteorder='little', signed = False)
		t += self.Reserved.to_bytes(4, byteorder='little', signed = False)
		t += self.CreationTime.to_bytes(8, byteorder='little', signed = False)
		t += self.LastAccessTime.to_bytes(8, byteorder='little', signed = False)
		t += self.LastWriteTime.to_bytes(8, byteorder='little', signed = False)
		t += self.ChangeTime.to_bytes(8, byteorder='little', signed = False)
		t += self.AllocationSize.to_bytes(8, byteorder='little', signed = False)
		t += self.EndofFile.to_bytes(8, byteorder='little', signed = False)
		t += self.FileAttributes.to_bytes(4, byteorder='little', signed = False)
		return t

	@staticmethod
	def from_bytes(bbuff):
		return CLOSE_REPLY.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		msg = CLOSE_REPLY()
		msg.StructureSize   = int.from_bytes(buff.read(2), byteorder='little')
		assert msg.StructureSize == 60
		msg.Flags  = CloseFlag(int.from_bytes(buff.read(2), byteorder='little'))
		msg.Reserved  = int.from_bytes(buff.read(4), byteorder='little')
		msg.CreationTime  = int.from_bytes(buff.read(8), byteorder='little')
		msg.LastAccessTime  = int.from_bytes(buff.read(8), byteorder='little')
		msg.LastWriteTime  = int.from_bytes(buff.read(8), byteorder='little')
		msg.ChangeTime  = int.from_bytes(buff.read(8), byteorder='little')
		msg.AllocationSize  = int.from_bytes(buff.read(8), byteorder='little')
		msg.EndofFile  = int.from_bytes(buff.read(8), byteorder='little')
		msg.FileAttributes  = FileAttributes(int.from_bytes(buff.read(4), byteorder='little'))
		return msg

	def __repr__(self):
		t = '==== SMB2 CLOSE REPLY ====\r\n'
		t += 'StructureSize: %s\r\n' % self.StructureSize
		return t



import enum
import io

from aiosmb.protocol.smb2.command_codes import SMB2Command
from aiosmb.wintypes.ntstatus import NTStatus
from aiosmb.protocol.smb2.headers.common import SMB2HeaderFlag

# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-smb2/ea4560b7-90da-4803-82b5-344754b92a79

class SMB2Header_ASYNC:
	def __init__(self):
		self.ProtocolId    = b'\xFESMB'
		self.StructureSize = 64
		self.CreditCharge  = None
		self.Status        = None  # In a request, this field is interpreted in different ways depending on the SMB2 dialect.
		self.Command       = None
		self.Credit        = None
		self.Flags         = None
		self.NextCommand   = None
		self.MessageId     = None
		self.AsyncId       = None
		self.SessionId     = None
		self.Signature     = None

	@staticmethod
	def from_buffer(buff):
		hdr = SMB2Header_ASYNC()
		hdr.ProtocolId = buff.read(4)
		assert hdr.ProtocolId == b'\xFESMB'
		hdr.StructureSize = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		assert hdr.StructureSize == 64
		hdr.CreditCharge = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		hdr.Status = NTStatus(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		hdr.Command = SMB2Command(int.from_bytes(buff.read(2), byteorder='little', signed = False))
		hdr.Credit =  int.from_bytes(buff.read(2), byteorder='little', signed = False)
		hdr.Flags =  SMB2HeaderFlag(int.from_bytes(buff.read(4), byteorder='little', signed = False))
		hdr.NextCommand = int.from_bytes(buff.read(4), byteorder='little', signed = False)
		hdr.MessageId = int.from_bytes(buff.read(8), byteorder='little', signed = False)
		hdr.AsyncId = buff.read(8)
		hdr.SessionId = buff.read(8)
		hdr.Signature = buff.read(16)
		return hdr

	@staticmethod
	def construct(cmd, flags, msgid, Credit = 0, NextCommand=0, CreditCharge = 0, 
					Signature=b'\x00'*16,
					AsyncId=b'\x00'*8, SessionId = b'\x00'*8, 
					status = NTStatus.SUCCESS):
		hdr = SMB2Header_ASYNC()
		hdr.ProtocolId = b'\xFESMB'
		hdr.StructureSize = 64
		hdr.CreditCharge = CreditCharge
		hdr.Status = status
		hdr.Command = cmd
		hdr.Credit =  Credit
		hdr.Flags =  flags
		hdr.NextCommand = NextCommand
		hdr.MessageId = msgid
		hdr.AsyncId = AsyncId
		hdr.SessionId = SessionId
		hdr.Signature = Signature

		return hdr

	def to_bytes(self):
		t  = self.ProtocolId
		t += self.StructureSize.to_bytes(2, byteorder = 'little', signed=False)
		t += self.CreditCharge.to_bytes(2, byteorder = 'little', signed=False)
		t += self.Status.value.to_bytes(4, byteorder = 'little', signed=False)
		t += self.Command.value.to_bytes(2, byteorder = 'little', signed=False)
		t += self.Credit.to_bytes(2, byteorder = 'little', signed=False)
		t += self.Flags.to_bytes(4, byteorder = 'little', signed=False)
		t += self.NextCommand.to_bytes(4, byteorder = 'little', signed=False)
		t += self.MessageId.to_bytes(8, byteorder = 'little', signed=False)
		t += self.AsyncId
		t += self.SessionId
		t += self.Signature
		return t

	def __repr__(self):
		t = '===SMB2 HEADER ASYNC===\r\n'
		t += 'ProtocolId: %s\r\n' % self.ProtocolId
		t += 'StructureSize: %s\r\n' % self.StructureSize
		t += 'CreditCharge: %s\r\n' % self.CreditCharge
		t += 'Status: %s\r\n' % self.Status.name
		t += 'Command: %s\r\n' % self.Command.name
		t += 'Credit: %s\r\n' % self.Credit
		t += 'Flags: %s\r\n' % self.Flags
		t += 'NextCommand: %s\r\n' % self.NextCommand
		t += 'MessageId: %s\r\n' % self.MessageId
		t += 'AsyncId: %s\r\n' % self.AsyncId
		t += 'SessionId: %s\r\n' % self.SessionId
		t += 'Signature: %s\r\n' % self.Signature
		return t



import logging
from aiosmb import logger
import traceback

#from aiosmb.dtyp.constrcuted_security.guid import GUID
from aiosmb.dcerpc.v5.common.secrets import SMBUserSecrets
from aiosmb.wintypes.dtyp.structures.filetime import FILETIME
from aiosmb.wintypes.dtyp.constrcuted_security.sid import SID
from aiosmb.dcerpc.v5.dtypes import NULL
from aiosmb.dcerpc.v5.uuid import string_to_bin
from aiosmb.dcerpc.v5 import drsuapi, samr
from aiosmb.dcerpc.v5.interfaces.endpointmgr import EPM
from aiosmb.dcerpc.v5.interfaces.servicemanager import *
from aiosmb.dcerpc.v5.rpcrt import RPC_C_AUTHN_LEVEL_PKT_INTEGRITY, RPC_C_AUTHN_LEVEL_PKT_PRIVACY, DCERPCException, RPC_C_AUTHN_GSS_NEGOTIATE

from aiosmb.commons.utils.decorators import red, rr

class SMBDRSUAPI:
	def __init__(self, connection, domainname = None):
		self.connection = connection	
		self.domainname = domainname
		
		self.dce = None
		self.handle = None
		
		self.__NtdsDsaObjectGuid = None
		self.__ppartialAttrSet = None
		
		self.ATTRTYP_TO_ATTID = {
				'userPrincipalName': '1.2.840.113556.1.4.656',
				'sAMAccountName': '1.2.840.113556.1.4.221',
				'unicodePwd': '1.2.840.113556.1.4.90',
				'dBCSPwd': '1.2.840.113556.1.4.55',
				'ntPwdHistory': '1.2.840.113556.1.4.94',
				'lmPwdHistory': '1.2.840.113556.1.4.160',
				'supplementalCredentials': '1.2.840.113556.1.4.125',
				'objectSid': '1.2.840.113556.1.4.146',
				'pwdLastSet': '1.2.840.113556.1.4.96',
				'userAccountControl':'1.2.840.113556.1.4.8',
			}
			
		self.NAME_TO_ATTRTYP = {
			'userPrincipalName': 0x90290,
			'sAMAccountName': 0x900DD,
			'unicodePwd': 0x9005A,
			'dBCSPwd': 0x90037,
			'ntPwdHistory': 0x9005E,
			'lmPwdHistory': 0x900A0,
			'supplementalCredentials': 0x9007D,
			'objectSid': 0x90092,
			'userAccountControl':0x90008,
		}
		
		self.KERBEROS_TYPE = {
			1:'dec-cbc-crc',
			3:'des-cbc-md5',
			17:'aes128-cts-hmac-sha1-96',
			18:'aes256-cts-hmac-sha1-96',
			0xffffff74:'rc4_hmac',
		}
		
	async def __aenter__(self):
		return self
	
	@red
	async def __aexit__(self, exc_type, exc, traceback):
		await self.close()
		return True,None
		
	
	@red
	async def connect(self, open = False):
		epm = EPM(self.connection, protocol = 'ncacn_ip_tcp')
		await rr(epm.connect())
		stringBinding, _ = await rr(epm.map(drsuapi.MSRPC_UUID_DRSUAPI))
		self.dce = epm.get_connection_from_stringbinding(stringBinding)

		#the line below must be set!
		self.dce.set_auth_level(RPC_C_AUTHN_LEVEL_PKT_PRIVACY)

		await rr(self.dce.connect())
		
		if open == True:
			await rr(self.open())
		return True,None
	
	@red
	async def open(self):
		if not self.dce:
			await rr(self.connect())
		
		await rr(self.dce.bind(drsuapi.MSRPC_UUID_DRSUAPI))
		request = drsuapi.DRSBind()
		request['puuidClientDsa'] = drsuapi.NTDSAPI_CLIENT_GUID
		drs = drsuapi.DRS_EXTENSIONS_INT()
		drs['cb'] = len(drs) #- 4
		drs['dwFlags'] = drsuapi.DRS_EXT_GETCHGREQ_V6 | drsuapi.DRS_EXT_GETCHGREPLY_V6 | drsuapi.DRS_EXT_GETCHGREQ_V8 | \
						 drsuapi.DRS_EXT_STRONG_ENCRYPTION
		drs['SiteObjGuid'] = drsuapi.NULLGUID
		drs['Pid'] = 0
		drs['dwReplEpoch'] = 0
		drs['dwFlagsExt'] = 0
		drs['ConfigObjGUID'] = drsuapi.NULLGUID
		# I'm uber potential (c) Ben
		drs['dwExtCaps'] = 0xffffffff
		request['pextClient']['cb'] = len(drs)
		request['pextClient']['rgb'] = list(drs.getData())
		resp,_ = await rr(self.dce.request(request))
		
		# Let's dig into the answer to check the dwReplEpoch. This field should match the one we send as part of
		# DRSBind's DRS_EXTENSIONS_INT(). If not, it will fail later when trying to sync data.
		drsExtensionsInt = drsuapi.DRS_EXTENSIONS_INT()

		# If dwExtCaps is not included in the answer, let's just add it so we can unpack DRS_EXTENSIONS_INT right.
		ppextServer = b''.join(resp['ppextServer']['rgb']) + b'\x00' * (
		len(drsuapi.DRS_EXTENSIONS_INT()) - resp['ppextServer']['cb'])
		drsExtensionsInt.fromString(ppextServer)

		if drsExtensionsInt['dwReplEpoch'] != 0:
			# Different epoch, we have to call DRSBind again
			if logger.level == logging.DEBUG:
				logger.debug("DC's dwReplEpoch != 0, setting it to %d and calling DRSBind again" % drsExtensionsInt[
					'dwReplEpoch'])
			drs['dwReplEpoch'] = drsExtensionsInt['dwReplEpoch']
			request['pextClient']['cb'] = len(drs)
			request['pextClient']['rgb'] = list(drs.getData())
			resp,_ = await rr(self.dce.request(request))

		self.handle = resp['phDrs']

		# Now let's get the NtdsDsaObjectGuid UUID to use when querying NCChanges
		resp, _ = await rr(drsuapi.hDRSDomainControllerInfo(self.dce, self.handle, self.domainname, 2))
		if logger.level == logging.DEBUG:
			logger.debug('DRSDomainControllerInfo() answer %s' % resp.dump())

		if resp['pmsgOut']['V2']['cItems'] > 0:
			self.__NtdsDsaObjectGuid = resp['pmsgOut']['V2']['rItems'][0]['NtdsDsaObjectGuid']
		else:
			logger.error("Couldn't get DC info for domain %s" % self.domainname)
			raise Exception('Fatal, aborting!')

		return True,None
	
	@red
	async def get_user_secrets(self, username):
		ra = {
			'userPrincipalName': '1.2.840.113556.1.4.656',
			'sAMAccountName': '1.2.840.113556.1.4.221',
			'unicodePwd': '1.2.840.113556.1.4.90',
			'dBCSPwd': '1.2.840.113556.1.4.55',
			'ntPwdHistory': '1.2.840.113556.1.4.94',
			'lmPwdHistory': '1.2.840.113556.1.4.160',
			'supplementalCredentials': '1.2.840.113556.1.4.125',
			'objectSid': '1.2.840.113556.1.4.146',
			'pwdLastSet': '1.2.840.113556.1.4.96',
			'userAccountControl':'1.2.840.113556.1.4.8'
		}
		formatOffered = drsuapi.DS_NT4_ACCOUNT_NAME_SANS_DOMAIN
		
		crackedName, _ = await rr(
			self.DRSCrackNames(
				formatOffered,
				drsuapi.DS_NAME_FORMAT.DS_UNIQUE_ID_NAME,
				name=username
			)
		)
		
		###### TODO: CHECKS HERE
		
		#guid = GUID.from_string(crackedName['pmsgOut']['V1']['pResult']['rItems'][0]['pName'][:-1][1:-1])
		guid = crackedName['pmsgOut']['V1']['pResult']['rItems'][0]['pName'][:-1][1:-1]
		
		userRecord, _ = await rr(self.DRSGetNCChanges(guid, ra))
		
		replyVersion = 'V%d' % userRecord['pdwOutVersion']
		if userRecord['pmsgOut'][replyVersion]['cNumObjects'] == 0:
			raise Exception('DRSGetNCChanges didn\'t return any object!')
		
		#print(userRecord.dump())
		#print(userRecord['pmsgOut'][replyVersion]['PrefixTableSrc']['pPrefixEntry'])
		
		record = userRecord
		prefixTable = userRecord['pmsgOut'][replyVersion]['PrefixTableSrc']['pPrefixEntry']
		##### decryption!
		logger.debug('Decrypting hash for user: %s' % record['pmsgOut'][replyVersion]['pNC']['StringName'][:-1])
		
		us = SMBUserSecrets()
		user_properties = None

		rid = int.from_bytes(record['pmsgOut'][replyVersion]['pObjects']['Entinf']['pName']['Sid'][-4:], 'little', signed = False)
		
		for attr in record['pmsgOut'][replyVersion]['pObjects']['Entinf']['AttrBlock']['pAttr']:
		
			try:
				attId = drsuapi.OidFromAttid(prefixTable, attr['attrTyp'])
				LOOKUP_TABLE = self.ATTRTYP_TO_ATTID
			except Exception as e:
				logger.error('Failed to execute OidFromAttid with error %s, fallbacking to fixed table' % e)
				logger.error('Exception', exc_info=True)
				# Fallbacking to fixed table and hope for the best
				attId = attr['attrTyp']
				LOOKUP_TABLE = self.NAME_TO_ATTRTYP
				
			if attId == LOOKUP_TABLE['dBCSPwd']:
				if attr['AttrVal']['valCount'] > 0:
					encrypteddBCSPwd = b''.join(attr['AttrVal']['pAVal'][0]['pVal'])
					encryptedLMHash = drsuapi.DecryptAttributeValue(self.dce.get_session_key(), encrypteddBCSPwd)
					us.lm_hash = drsuapi.removeDESLayer(encryptedLMHash, rid)
				else:
					us.lm_hash = bytes.fromhex('aad3b435b51404eeaad3b435b51404ee')
					
			elif attId == LOOKUP_TABLE['unicodePwd']:
				if attr['AttrVal']['valCount'] > 0:
					encryptedUnicodePwd = b''.join(attr['AttrVal']['pAVal'][0]['pVal'])
					encryptedNTHash = drsuapi.DecryptAttributeValue(self.dce.get_session_key(), encryptedUnicodePwd)
					us.nt_hash = drsuapi.removeDESLayer(encryptedNTHash, rid)
				else:
					us.nt_hash = bytes.fromhex('31d6cfe0d16ae931b73c59d7e0c089c0')
					
			elif attId == LOOKUP_TABLE['userPrincipalName']:
				if attr['AttrVal']['valCount'] > 0:
					try:
						us.domain = b''.join(attr['AttrVal']['pAVal'][0]['pVal']).decode('utf-16le').split('@')[-1]
					except:
						us.domain = None
				else:
					us.domain = None
						
			elif attId == LOOKUP_TABLE['sAMAccountName']:
				if attr['AttrVal']['valCount'] > 0:
					try:
						us.username = b''.join(attr['AttrVal']['pAVal'][0]['pVal']).decode('utf-16le')
					except Exception as e:
						logger.error('Cannot get sAMAccountName for %s' % record['pmsgOut'][replyVersion]['pNC']['StringName'][:-1])
						us.username = 'unknown'
				else:
					logger.error('Cannot get sAMAccountName for %s' % record['pmsgOut'][replyVersion]['pNC']['StringName'][:-1])
					us.username = 'unknown'
						
			elif attId == LOOKUP_TABLE['objectSid']:
				if attr['AttrVal']['valCount'] > 0:
					us.object_sid = SID.from_bytes(b''.join(attr['AttrVal']['pAVal'][0]['pVal']))
				else:
					logger.error('Cannot get objectSid for %s' % record['pmsgOut'][replyVersion]['pNC']['StringName'][:-1])
					us.object_sid = rid
			elif attId == LOOKUP_TABLE['pwdLastSet']:
				if attr['AttrVal']['valCount'] > 0:
					try:
						
						us.pwd_last_set = FILETIME.from_bytes(b''.join(attr['AttrVal']['pAVal'][0]['pVal'])).datetime.isoformat()
					except Exception as e:
						
						logger.error('Cannot get pwdLastSet for %s' % record['pmsgOut'][replyVersion]['pNC']['StringName'][:-1])
						us.pwd_last_set = None
						
			elif attId == LOOKUP_TABLE['userAccountControl']:
				if attr['AttrVal']['valCount'] > 0:
					us.user_account_status = int.from_bytes(b''.join(attr['AttrVal']['pAVal'][0]['pVal']), 'little', signed = False)
				else:
					us.user_account_status = None
					
			if attId == LOOKUP_TABLE['lmPwdHistory']:
				if attr['AttrVal']['valCount'] > 0:
					encryptedLMHistory = b''.join(attr['AttrVal']['pAVal'][0]['pVal'])
					tmpLMHistory = drsuapi.DecryptAttributeValue(self.dce.get_session_key(), encryptedLMHistory)
					for i in range(0, len(tmpLMHistory) // 16):
						LMHashHistory = drsuapi.removeDESLayer(tmpLMHistory[i * 16:(i + 1) * 16], rid)
						us.lm_history.append(LMHashHistory)
				else:
					logger.debug('No lmPwdHistory for user %s' % record['pmsgOut'][replyVersion]['pNC']['StringName'][:-1])
			elif attId == LOOKUP_TABLE['ntPwdHistory']:
				if attr['AttrVal']['valCount'] > 0:
					encryptedNTHistory = b''.join(attr['AttrVal']['pAVal'][0]['pVal'])
					tmpNTHistory = drsuapi.DecryptAttributeValue(self.dce.get_session_key(), encryptedNTHistory)
					for i in range(0, len(tmpNTHistory) // 16):
						NTHashHistory = drsuapi.removeDESLayer(tmpNTHistory[i * 16:(i + 1) * 16], rid)
						us.nt_history.append(NTHashHistory)
				else:
					logger.debug('No ntPwdHistory for user %s' % record['pmsgOut'][replyVersion]['pNC']['StringName'][:-1])
					
			elif attId == LOOKUP_TABLE['supplementalCredentials']:
				if attr['AttrVal']['valCount'] > 0:
					blob = b''.join(attr['AttrVal']['pAVal'][0]['pVal'])
					supplementalCredentials = drsuapi.DecryptAttributeValue(self.dce.get_session_key(), blob)
					if len(supplementalCredentials) < 24:
						supplementalCredentials = None
						
					else:
						try:
							user_properties = samr.USER_PROPERTIES(supplementalCredentials)
						except Exception as e:
							# On some old w2k3 there might be user properties that don't
							# match [MS-SAMR] structure, discarding them
							pass
			
		
		if user_properties is not None:
			propertiesData = user_properties['UserProperties']
			for propertyCount in range(user_properties['PropertyCount']):
				userProperty = samr.USER_PROPERTY(propertiesData)
				propertiesData = propertiesData[len(userProperty):]
				# For now, we will only process Newer Kerberos Keys and CLEARTEXT
				if userProperty['PropertyName'].decode('utf-16le') == 'Primary:Kerberos-Newer-Keys':
					propertyValueBuffer = bytes.fromhex(userProperty['PropertyValue'].decode())
					kerbStoredCredentialNew = samr.KERB_STORED_CREDENTIAL_NEW(propertyValueBuffer)
					data = kerbStoredCredentialNew['Buffer']
					for credential in range(kerbStoredCredentialNew['CredentialCount']):
						keyDataNew = samr.KERB_KEY_DATA_NEW(data)
						data = data[len(keyDataNew):]
						keyValue = propertyValueBuffer[keyDataNew['KeyOffset']:][:keyDataNew['KeyLength']]

						if  keyDataNew['KeyType'] in self.KERBEROS_TYPE:
							answer =  (self.KERBEROS_TYPE[keyDataNew['KeyType']],keyValue)
						else:
							answer =  (hex(keyDataNew['KeyType']),keyValue)
						# We're just storing the keys, not printing them, to make the output more readable
						# This is kind of ugly... but it's what I came up with tonight to get an ordered
						# set :P. Better ideas welcomed ;)
						us.kerberos_keys.append(answer)
				elif userProperty['PropertyName'].decode('utf-16le') == 'Primary:CLEARTEXT':
					# [MS-SAMR] 3.1.1.8.11.5 Primary:CLEARTEXT Property
					# This credential type is the cleartext password. The value format is the UTF-16 encoded cleartext password.
					# SkelSec: well, almost. actually the property is the hex-encoded bytes of an UTF-16LE encoded plaintext string
					encoded_pw = bytes.fromhex(userProperty['PropertyValue'].decode('ascii'))
					try:
						answer = encoded_pw.decode('utf-16le')
					except UnicodeDecodeError:
						# This could be because we're decoding a machine password. Printing it hex
						answer = encoded_pw.decode('utf-8')

					us.cleartext_pwds.append(answer)
			
		
		return us, None
			
	@red
	async def DRSCrackNames(self, formatOffered=drsuapi.DS_NAME_FORMAT.DS_DISPLAY_NAME, formatDesired=drsuapi.DS_NAME_FORMAT.DS_FQDN_1779_NAME, name=''):
		if self.handle is None:
			await rr(self.open())

		logger.debug('Calling DRSCrackNames for %s' % name)
		resp, _ = await rr(drsuapi.hDRSCrackNames(self.dce, self.handle, 0, formatOffered, formatDesired, (name,)))
		return resp, None
	
	@red
	async def DRSGetNCChanges(self, guid, req_attributes = {}):
		if self.handle is None:
			await rr(self.open())

		logger.debug('Calling DRSGetNCChanges for %s ' % guid)
		request = drsuapi.DRSGetNCChanges()
		request['hDrs'] = self.handle
		request['dwInVersion'] = 8

		request['pmsgIn']['tag'] = 8
		request['pmsgIn']['V8']['uuidDsaObjDest'] = self.__NtdsDsaObjectGuid
		request['pmsgIn']['V8']['uuidInvocIdSrc'] = self.__NtdsDsaObjectGuid

		dsName = drsuapi.DSNAME()
		dsName['SidLen'] = 0
		dsName['Guid'] = string_to_bin(guid)#guid.to_bytes()
		dsName['Sid'] = ''
		dsName['NameLen'] = 0
		dsName['StringName'] = ('\x00')

		dsName['structLen'] = len(dsName.getData())

		request['pmsgIn']['V8']['pNC'] = dsName

		request['pmsgIn']['V8']['usnvecFrom']['usnHighObjUpdate'] = 0
		request['pmsgIn']['V8']['usnvecFrom']['usnHighPropUpdate'] = 0

		request['pmsgIn']['V8']['pUpToDateVecDest'] = NULL

		request['pmsgIn']['V8']['ulFlags'] =  drsuapi.DRS_INIT_SYNC | drsuapi.DRS_WRIT_REP
		request['pmsgIn']['V8']['cMaxObjects'] = 1
		request['pmsgIn']['V8']['cMaxBytes'] = 0
		request['pmsgIn']['V8']['ulExtendedOp'] = drsuapi.EXOP_REPL_OBJ
		if self.__ppartialAttrSet is None:
			self.__prefixTable = []
			self.__ppartialAttrSet = drsuapi.PARTIAL_ATTR_VECTOR_V1_EXT()
			self.__ppartialAttrSet['dwVersion'] = 1
			self.__ppartialAttrSet['cAttrs'] = len(req_attributes)
			for attId in list(req_attributes.values()):
				self.__ppartialAttrSet['rgPartialAttr'].append(drsuapi.MakeAttid(self.__prefixTable , attId))
		request['pmsgIn']['V8']['pPartialAttrSet'] = self.__ppartialAttrSet
		request['pmsgIn']['V8']['PrefixTableDest']['PrefixCount'] = len(self.__prefixTable)
		request['pmsgIn']['V8']['PrefixTableDest']['pPrefixEntry'] = self.__prefixTable
		request['pmsgIn']['V8']['pPartialAttrSetEx1'] = NULL

		data, _ = await rr(self.dce.request(request))
		return data, None
	
	@red
	async def close(self):
		if self.handle:
			try:
				await rr(drsuapi.hDRSUnbind(self.dce, self.handle))
			except:
				pass
		if self.dce:
			try:
				await rr(self.dce.disconnect())
			except:
				pass
		
		return True,None


import io

# https://msdn.microsoft.com/en-us/library/ee441913.aspx
class SMB_COM_NEGOTIATE_REQ:
	def __init__(self):
		##### parameters ####
		self.WordCount = 0
		##### SMB_Data ###
		self.ByteCount = None
		self.Dialects  = [] #list fo dialect strings

	@staticmethod
	def from_bytes(bbuff):
		return SMB_COM_NEGOTIATE_REQ.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		cmd = SMB_COM_NEGOTIATE_REQ()
		cmd.WordCount = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		cmd.ByteCount = int.from_bytes(buff.read(2), byteorder='little', signed = False)
		dialect_buffer = buff.read(cmd.ByteCount)
		#print(dialect_buffer)
		
		i = 0
		while i < len(dialect_buffer):
			if not dialect_buffer[i] == 2:
				#print(i)
				raise Exception('Dialect buffer expected byte 0x02!')
			m = dialect_buffer[i:].find(b'\x00')
			if m == -1:
				raise Exception('Could not find end of string!')
			cmd.Dialects.append(dialect_buffer[i+1: m+i].decode('ascii'))
			i += m+1
		return cmd
		
		
	def to_bytes(self):
		t = self.WordCount.to_bytes(1, byteorder='little', signed = False)
		
		dialect_buffer = b''
		for dialect in self.Dialects:
			dialect_buffer += b'\x02' + dialect.encode('ascii') + b'\x00'
			
		t+= len(dialect_buffer).to_bytes(2, byteorder='little', signed = False)
		t+= dialect_buffer
		
		return t

	def __repr__(self):
		t = '===SMB_COM_NEGOTIATE_REQ===\r\n'
		return t
		
def test():
	data = bytes.fromhex('002200024e54204c4d20302e31320002534d4220322e3030320002534d4220322e3f3f3f00')
	nego = SMB_COM_NEGOTIATE_REQ.from_bytes(data)
	print(nego.Dialects)
	if 'NT LM 0.12' not in nego.Dialects:
		raise Exception('Error!')
	if 'SMB 2.002' not in nego.Dialects:
		raise Exception('Error!')
	if 'SMB 2.???' not in nego.Dialects:
		raise Exception('Error!')
		
	data_reconstructed = nego.to_bytes()
	assert data == data_reconstructed
		
if __name__ == '__main__':
	test()


# SECUREAUTH LABS. Copyright 2018 SecureAuth Corporation. All rights reserved.
#
# This software is provided under under a slightly modified version
# of the Apache Software License. See the accompanying LICENSE file
# for more information.
#
# Author: Alberto Solino (@agsolino)
#
# Description:
#   [MS-RPRN] Interface implementation
#
#   Best way to learn how to use these calls is to grab the protocol standard
#   so you understand what the call does, and then read the test case located
#   at https://github.com/SecureAuthCorp/impacket/tree/master/tests/SMB_RPC
#
#   Some calls have helper functions, which makes it even easier to use.
#   They are located at the end of this file. 
#   Helper functions start with "h"<name of the call>.
#   There are test cases for them too. 
#

from aiosmb.dcerpc.v5.dtypes import ULONGLONG, UINT, USHORT, LPWSTR, DWORD, ULONG, NULL
from aiosmb.dcerpc.v5.ndr import NDRCALL, NDRSTRUCT, NDRUNION, NDRPOINTER, NDRUniConformantArray
from aiosmb.dcerpc.v5.rpcrt import DCERPCException
from aiosmb.dcerpc.v5 import system_errors
from aiosmb.dcerpc.v5.uuid import uuidtup_to_bin

MSRPC_UUID_RPRN = uuidtup_to_bin(('12345678-1234-ABCD-EF00-0123456789AB', '1.0'))

class DCERPCSessionError(DCERPCException):
	def __init__(self, error_string=None, error_code=None, packet=None):
		DCERPCException.__init__(self, error_string, error_code, packet)

	def __str__( self ):
		key = self.error_code
		if key in system_errors.ERROR_MESSAGES:
			error_msg_short = system_errors.ERROR_MESSAGES[key][0]
			error_msg_verbose = system_errors.ERROR_MESSAGES[key][1] 
			return 'RPRN SessionError: code: 0x%x - %s - %s' % (self.error_code, error_msg_short, error_msg_verbose)
		else:
			return 'RPRN SessionError: unknown error code: 0x%x' % self.error_code

################################################################################
# CONSTANTS
################################################################################
# 2.2.1.1.7 STRING_HANDLE
STRING_HANDLE = LPWSTR
class PSTRING_HANDLE(NDRPOINTER):
	referent = (
		('Data', STRING_HANDLE),
	)

# 2.2.3.1 Access Values
JOB_ACCESS_ADMINISTER         = 0x00000010
JOB_ACCESS_READ               = 0x00000020
JOB_EXECUTE                   = 0x00020010
JOB_READ                      = 0x00020020
JOB_WRITE                     = 0x00020010
JOB_ALL_ACCESS                = 0x000F0030
PRINTER_ACCESS_ADMINISTER     = 0x00000004
PRINTER_ACCESS_USE            = 0x00000008
PRINTER_ACCESS_MANAGE_LIMITED = 0x00000040
PRINTER_ALL_ACCESS            = 0x000F000C
PRINTER_EXECUTE               = 0x00020008
PRINTER_READ                  = 0x00020008
PRINTER_WRITE                 = 0x00020008
SERVER_ACCESS_ADMINISTER      = 0x00000001
SERVER_ACCESS_ENUMERATE       = 0x00000002
SERVER_ALL_ACCESS             = 0x000F0003
SERVER_EXECUTE                = 0x00020002
SERVER_READ                   = 0x00020002
SERVER_WRITE                  = 0x00020003
SPECIFIC_RIGHTS_ALL           = 0x0000FFFF
STANDARD_RIGHTS_ALL           = 0x001F0000
STANDARD_RIGHTS_EXECUTE       = 0x00020000
STANDARD_RIGHTS_READ          = 0x00020000
STANDARD_RIGHTS_REQUIRED      = 0x000F0000
STANDARD_RIGHTS_WRITE         = 0x00020000
SYNCHRONIZE                   = 0x00100000
DELETE                        = 0x00010000
READ_CONTROL                  = 0x00020000
WRITE_DAC                     = 0x00040000
WRITE_OWNER                   = 0x00080000
GENERIC_READ                  = 0x80000000
GENERIC_WRITE                 = 0x40000000
GENERIC_EXECUTE               = 0x20000000
GENERIC_ALL                   = 0x10000000

# 2.2.3.6.1 Printer Change Flags for Use with a Printer Handle
PRINTER_CHANGE_SET_PRINTER        = 0x00000002
PRINTER_CHANGE_DELETE_PRINTER     = 0x00000004
PRINTER_CHANGE_PRINTER            = 0x000000FF
PRINTER_CHANGE_ADD_JOB            = 0x00000100
PRINTER_CHANGE_SET_JOB            = 0x00000200
PRINTER_CHANGE_DELETE_JOB         = 0x00000400
PRINTER_CHANGE_WRITE_JOB          = 0x00000800
PRINTER_CHANGE_JOB                = 0x0000FF00
PRINTER_CHANGE_SET_PRINTER_DRIVER = 0x20000000
PRINTER_CHANGE_TIMEOUT            = 0x80000000
PRINTER_CHANGE_ALL                = 0x7777FFFF
PRINTER_CHANGE_ALL_2              = 0x7F77FFFF

# 2.2.3.6.2 Printer Change Flags for Use with a Server Handle
PRINTER_CHANGE_ADD_PRINTER_DRIVER        = 0x10000000
PRINTER_CHANGE_DELETE_PRINTER_DRIVER     = 0x40000000
PRINTER_CHANGE_PRINTER_DRIVER            = 0x70000000
PRINTER_CHANGE_ADD_FORM                  = 0x00010000
PRINTER_CHANGE_DELETE_FORM               = 0x00040000
PRINTER_CHANGE_SET_FORM                  = 0x00020000
PRINTER_CHANGE_FORM                      = 0x00070000
PRINTER_CHANGE_ADD_PORT                  = 0x00100000
PRINTER_CHANGE_CONFIGURE_PORT            = 0x00200000
PRINTER_CHANGE_DELETE_PORT               = 0x00400000
PRINTER_CHANGE_PORT                      = 0x00700000
PRINTER_CHANGE_ADD_PRINT_PROCESSOR       = 0x01000000
PRINTER_CHANGE_DELETE_PRINT_PROCESSOR    = 0x04000000
PRINTER_CHANGE_PRINT_PROCESSOR           = 0x07000000
PRINTER_CHANGE_ADD_PRINTER               = 0x00000001
PRINTER_CHANGE_FAILED_CONNECTION_PRINTER = 0x00000008
PRINTER_CHANGE_SERVER                    = 0x08000000

# 2.2.3.7 Printer Enumeration Flags
PRINTER_ENUM_LOCAL       = 0x00000002
PRINTER_ENUM_CONNECTIONS = 0x00000004
PRINTER_ENUM_NAME        = 0x00000008
PRINTER_ENUM_REMOTE      = 0x00000010
PRINTER_ENUM_SHARED      = 0x00000020
PRINTER_ENUM_NETWORK     = 0x00000040
PRINTER_ENUM_EXPAND      = 0x00004000
PRINTER_ENUM_CONTAINER   = 0x00008000
PRINTER_ENUM_ICON1       = 0x00010000
PRINTER_ENUM_ICON2       = 0x00020000
PRINTER_ENUM_ICON3       = 0x00040000
PRINTER_ENUM_ICON8       = 0x00800000
PRINTER_ENUM_HIDE        = 0x01000000


# 2.2.3.8 Printer Notification Values
PRINTER_NOTIFY_CATEGORY_2D  = 0x00000000
PRINTER_NOTIFY_CATEGORY_ALL = 0x00010000
PRINTER_NOTIFY_CATEGORY_3D  = 0x00020000


################################################################################
# STRUCTURES
################################################################################
# 2.2.1.1.4 PRINTER_HANDLE
class PRINTER_HANDLE(NDRSTRUCT):
	structure =  (
		('Data','20s=b""'),
	)
	def getAlignment(self):
		if self._isNDR64 is True:
			return 8
		else:
			return 4

# 2.2.1.2.1 DEVMODE_CONTAINER
class BYTE_ARRAY(NDRUniConformantArray):
	item = 'c'

class PBYTE_ARRAY(NDRPOINTER):
	referent = (
		('Data', BYTE_ARRAY),
	)

class DEVMODE_CONTAINER(NDRSTRUCT):
	structure =  (
		('cbBuf',DWORD),
		('pDevMode',PBYTE_ARRAY),
	)

# 2.2.1.11.1 SPLCLIENT_INFO_1
class SPLCLIENT_INFO_1(NDRSTRUCT):
	structure =  (
		('dwSize',DWORD),
		('pMachineName',LPWSTR),
		('pUserName',LPWSTR),
		('dwBuildNum',DWORD),
		('dwMajorVersion',DWORD),
		('dwMinorVersion',DWORD),
		('wProcessorArchitecture',USHORT),
	)

class PSPLCLIENT_INFO_1(NDRPOINTER):
	referent = (
		('Data', SPLCLIENT_INFO_1),
	)

# 2.2.1.11.2 SPLCLIENT_INFO_2
class SPLCLIENT_INFO_2(NDRSTRUCT):
	structure =  (
		('notUsed',ULONGLONG),
	)

class PSPLCLIENT_INFO_2(NDRPOINTER):
	referent = (
		('Data', SPLCLIENT_INFO_2),
	)
# 2.2.1.11.3 SPLCLIENT_INFO_3
class SPLCLIENT_INFO_3(NDRSTRUCT):
	structure =  (
		('cbSize',UINT),
		('dwFlags',DWORD),
		('dwFlags',DWORD),
		('pMachineName',LPWSTR),
		('pUserName',LPWSTR),
		('dwBuildNum',DWORD),
		('dwMajorVersion',DWORD),
		('dwMinorVersion',DWORD),
		('wProcessorArchitecture',USHORT),
		('hSplPrinter',ULONGLONG),
	)

class PSPLCLIENT_INFO_3(NDRPOINTER):
	referent = (
		('Data', SPLCLIENT_INFO_3),
	)
# 2.2.1.2.14 SPLCLIENT_CONTAINER
class CLIENT_INFO_UNION(NDRUNION):
	commonHdr = (
		('tag', ULONG),
	)
	union = {
		1 : ('pClientInfo1', PSPLCLIENT_INFO_1),
		2 : ('pNotUsed1', PSPLCLIENT_INFO_2),
		3 : ('pNotUsed2', PSPLCLIENT_INFO_3),
	}

class SPLCLIENT_CONTAINER(NDRSTRUCT):
	structure =  (
		('Level',DWORD),
		('ClientInfo',CLIENT_INFO_UNION),
	)


# 2.2.1.13.2 RPC_V2_NOTIFY_OPTIONS_TYPE
class USHORT_ARRAY(NDRUniConformantArray):
	item = '<H'

class PUSHORT_ARRAY(NDRPOINTER):
	referent = (
		('Data', USHORT_ARRAY),
	)

class RPC_V2_NOTIFY_OPTIONS_TYPE(NDRSTRUCT):
	structure =  (
		('Type',USHORT),
		('Reserved0',USHORT),
		('Reserved1',DWORD),
		('Reserved2',DWORD),
		('Count',DWORD),
		('pFields',PUSHORT_ARRAY),
	)

class PRPC_V2_NOTIFY_OPTIONS_TYPE_ARRAY(NDRPOINTER):
	referent = (
		('Data', RPC_V2_NOTIFY_OPTIONS_TYPE),
	)

# 2.2.1.13.1 RPC_V2_NOTIFY_OPTIONS
class RPC_V2_NOTIFY_OPTIONS(NDRSTRUCT):
	structure =  (
		('Version',DWORD),
		('Reserved',DWORD),
		('Count',DWORD),
		('pTypes',PRPC_V2_NOTIFY_OPTIONS_TYPE_ARRAY),
	)

class PRPC_V2_NOTIFY_OPTIONS(NDRPOINTER):
	referent = (
		('Data', RPC_V2_NOTIFY_OPTIONS),
	)


################################################################################
# RPC CALLS
################################################################################
# 3.1.4.2.1 RpcEnumPrinters (Opnum 0)
class RpcEnumPrinters(NDRCALL):
	opnum = 0
	structure = (
	   ('Flags', DWORD),
	   ('Name', STRING_HANDLE),
	   ('Level', DWORD),
	   ('pPrinterEnum', PBYTE_ARRAY),
	   ('cbBuf', DWORD),
	)

class RpcEnumPrintersResponse(NDRCALL):
	structure = (
	   ('pPrinterEnum', PBYTE_ARRAY),
	   ('pcbNeeded', DWORD),
	   ('pcReturned', DWORD),
	   ('ErrorCode', ULONG),
	)
# 3.1.4.2.2 RpcOpenPrinter (Opnum 1)
class RpcOpenPrinter(NDRCALL):
	opnum = 1
	structure = (
	   ('pPrinterName', STRING_HANDLE),
	   ('pDatatype', LPWSTR),
	   ('pDevModeContainer', DEVMODE_CONTAINER),
	   ('AccessRequired', DWORD),
	)

class RpcOpenPrinterResponse(NDRCALL):
	structure = (
	   ('pHandle', PRINTER_HANDLE),
	   ('ErrorCode', ULONG),
	)

# 3.1.4.2.9 RpcClosePrinter (Opnum 29)
class RpcClosePrinter(NDRCALL):
	opnum = 29
	structure = (
	   ('phPrinter', PRINTER_HANDLE),
	)

class RpcClosePrinterResponse(NDRCALL):
	structure = (
	   ('phPrinter', PRINTER_HANDLE),
	   ('ErrorCode', ULONG),
	)

# 3.1.4.10.4 RpcRemoteFindFirstPrinterChangeNotificationEx (Opnum 65)
class RpcRemoteFindFirstPrinterChangeNotificationEx(NDRCALL):
	opnum = 65
	structure = (
	   ('hPrinter', PRINTER_HANDLE),
	   ('fdwFlags', DWORD),
	   ('fdwOptions', DWORD),
	   ('pszLocalMachine', LPWSTR),
	   ('dwPrinterLocal', DWORD),
	   ('pOptions', PRPC_V2_NOTIFY_OPTIONS),
	)

class RpcRemoteFindFirstPrinterChangeNotificationExResponse(NDRCALL):
	structure = (
	   ('ErrorCode', ULONG),
	)

# 3.1.4.2.14 RpcOpenPrinterEx (Opnum 69)
class RpcOpenPrinterEx(NDRCALL):
	opnum = 69
	structure = (
	   ('pPrinterName', STRING_HANDLE),
	   ('pDatatype', LPWSTR),
	   ('pDevModeContainer', DEVMODE_CONTAINER),
	   ('AccessRequired', DWORD),
	   ('pClientInfo', SPLCLIENT_CONTAINER),
	)

class RpcOpenPrinterExResponse(NDRCALL):
	structure = (
	   ('pHandle', PRINTER_HANDLE),
	   ('ErrorCode', ULONG),
	)

################################################################################
# OPNUMs and their corresponding structures
################################################################################
OPNUMS = {
	0  : (RpcEnumPrinters, RpcEnumPrintersResponse),
	1  : (RpcOpenPrinter, RpcOpenPrinterResponse),
	29 : (RpcClosePrinter, RpcClosePrinterResponse),
	65 : (RpcRemoteFindFirstPrinterChangeNotificationEx, RpcRemoteFindFirstPrinterChangeNotificationExResponse),
	69 : (RpcOpenPrinterEx, RpcOpenPrinterExResponse),
}

################################################################################
# HELPER FUNCTIONS
################################################################################
def checkNullString(string):
	if string == NULL:
		return string

	if string[-1:] != '\x00':
		return string + '\x00'
	else:
		return string

async def hRpcOpenPrinter(dce, printerName, pDatatype = NULL, pDevModeContainer = NULL, accessRequired = SERVER_READ):
	"""
	RpcOpenPrinter retrieves a handle for a printer, port, port monitor, print job, or print server.
	Full Documentation: https://msdn.microsoft.com/en-us/library/cc244808.aspx

	:param DCERPC_v5 dce: a connected DCE instance.
	:param string printerName: A string for a printer connection, printer object, server object, job object, port
	object, or port monitor object. This MUST be a Domain Name System (DNS), NetBIOS, Internet Protocol version 4
	(IPv4), Internet Protocol version 6 (IPv6), or Universal Naming Convention (UNC) name that remote procedure
	call (RPC) binds to, and it MUST uniquely identify a print server on the network.
	:param string pDatatype: A string that specifies the data type to be associated with the printer handle.
	:param DEVMODE_CONTAINER pDevModeContainer: A DEVMODE_CONTAINER structure. This parameter MUST adhere to the specification in
	DEVMODE_CONTAINER Parameters (section 3.1.4.1.8.1).
	:param int accessRequired: The access level that the client requires for interacting with the object to which a
	handle is being opened.

	:return: a RpcOpenPrinterResponse instance, raises DCERPCSessionError on error.
	"""
	request = RpcOpenPrinter()
	request['pPrinterName'] = checkNullString(printerName)
	request['pDatatype'] = pDatatype
	if pDevModeContainer is NULL:
		request['pDevModeContainer']['pDevMode'] = NULL
	else:
		request['pDevModeContainer'] = pDevModeContainer

	request['AccessRequired'] = accessRequired
	return await dce.request(request)

async def hRpcClosePrinter(dce, phPrinter):
	"""
	RpcClosePrinter closes a handle to a printer object, server object, job object, or port object.
	Full Documentation: https://msdn.microsoft.com/en-us/library/cc244768.aspx

	:param DCERPC_v5 dce: a connected DCE instance.
	:param PRINTER_HANDLE phPrinter: A handle to a printer object, server object, job object, or port object.

	:return: a RpcClosePrinterResponse instance, raises DCERPCSessionError on error.
	"""
	request = RpcClosePrinter()
	request['phPrinter'] = phPrinter
	return await dce.request(request)


async def hRpcOpenPrinterEx(dce, printerName, pDatatype=NULL, pDevModeContainer=NULL, accessRequired=SERVER_READ,
					  pClientInfo=NULL):
	"""
	RpcOpenPrinterEx retrieves a handle for a printer, port, port monitor, print job, or print server
	Full Documentation: https://msdn.microsoft.com/en-us/library/cc244809.aspx

	:param DCERPC_v5 dce: a connected DCE instance.
	:param string printerName: A string for a printer connection, printer object, server object, job object, port
	object, or port monitor object. This MUST be a Domain Name System (DNS), NetBIOS, Internet Protocol version 4
	(IPv4), Internet Protocol version 6 (IPv6), or Universal Naming Convention (UNC) name that remote procedure
	call (RPC) binds to, and it MUST uniquely identify a print server on the network.
	:param string pDatatype: A string that specifies the data type to be associated with the printer handle.
	:param DEVMODE_CONTAINER pDevModeContainer: A DEVMODE_CONTAINER structure. This parameter MUST adhere to the specification in
	DEVMODE_CONTAINER Parameters (section 3.1.4.1.8.1).
	:param int accessRequired: The access level that the client requires for interacting with the object to which a
	handle is being opened.
	:param SPLCLIENT_CONTAINER pClientInfo: This parameter MUST adhere to the specification in SPLCLIENT_CONTAINER Parameters.

	:return: a RpcOpenPrinterExResponse instance, raises DCERPCSessionError on error.
	"""
	request = RpcOpenPrinterEx()
	request['pPrinterName'] = checkNullString(printerName)
	request['pDatatype'] = pDatatype
	if pDevModeContainer is NULL:
		request['pDevModeContainer']['pDevMode'] = NULL
	else:
		request['pDevModeContainer'] = pDevModeContainer

	request['AccessRequired'] = accessRequired
	if pClientInfo is NULL:
		raise Exception('pClientInfo cannot be NULL')

	request['pClientInfo'] = pClientInfo
	return await dce.request(request)


async def hRpcRemoteFindFirstPrinterChangeNotificationEx(dce, hPrinter, fdwFlags, fdwOptions=0, pszLocalMachine=NULL,
												   dwPrinterLocal=0, pOptions=NULL):
	"""
	creates a remote change notification object that monitors changes to printer objects and sends change notifications
	to a print client using either RpcRouterReplyPrinter (section 3.2.4.1.2) or RpcRouterReplyPrinterEx (section 3.2.4.1.4)
	Full Documentation: https://msdn.microsoft.com/en-us/library/cc244813.aspx

	:param DCERPC_v5 dce: a connected DCE instance.
	:param PRINTER_HANDLE hPrinter: A handle to a printer or server object.
	:param int fdwFlags: Flags that specify the conditions that are required for a change notification object to enter a signaled state.
	:param int fdwOptions: The category of printers for which change notifications are returned.
	:param string pszLocalMachine: A string that represents the name of the client computer.
	:param int dwPrinterLocal: An implementation-specific unique value that MUST be sufficient for the client to determine
	whether a call to RpcReplyOpenPrinter by the server is associated with the hPrinter parameter in this call.
	:param RPC_V2_NOTIFY_OPTIONS pOptions:  An RPC_V2_NOTIFY_OPTIONS structure that specifies printer or job members that the client listens to for notifications.

	:return: a RpcRemoteFindFirstPrinterChangeNotificationExResponse instance, raises DCERPCSessionError on error.
	"""
	request = RpcRemoteFindFirstPrinterChangeNotificationEx()

	request['hPrinter'] = hPrinter
	request['fdwFlags'] = fdwFlags
	request['fdwOptions'] = fdwOptions
	request['dwPrinterLocal'] = dwPrinterLocal
	if pszLocalMachine is NULL:
		raise Exception('pszLocalMachine cannot be NULL')
	request['pszLocalMachine'] = checkNullString(pszLocalMachine)
	request['pOptions'] = pOptions
	return await dce.request(request)

async def hRpcEnumPrinters(dce, flags, name = NULL, level = 1):
	"""
	RpcEnumPrinters enumerates available printers, print servers, domains, or print providers.
	Full Documentation: https://msdn.microsoft.com/en-us/library/cc244794.aspx

	:param DCERPC_v5 dce: a connected DCE instance.
	:param int flags: The types of print objects that this method enumerates. The value of this parameter is the
	result of a bitwise OR of one or more of the Printer Enumeration Flags (section 2.2.3.7).
	:param string name: NULL or a server name parameter as specified in Printer Server Name Parameters (section 3.1.4.1.4).
	:param level: The level of printer information structure.

	:return: a RpcEnumPrintersResponse instance, raises DCERPCSessionError on error.
	"""
	request = RpcEnumPrinters()
	request['Flags'] = flags
	request['Name'] = name
	request['pPrinterEnum'] = NULL
	request['Level'] = level
	bytesNeeded = 0
	try:
		dce.request(request)
	except DCERPCSessionError as e:
		if str(e).find('ERROR_INSUFFICIENT_BUFFER') < 0:
			raise
		bytesNeeded = e.get_packet()['pcbNeeded']

	request = RpcEnumPrinters()
	request['Flags'] = flags
	request['Name'] = name
	request['Level'] = level

	request['cbBuf'] = bytesNeeded
	request['pPrinterEnum'] = b'a' * bytesNeeded
	return await dce.request(request)
